////////////////////////////////////////////////////////////////////////
/*
     FILE: AVHRRLocator.java
  PURPOSE: Computes AVHRR Earth locations.
   AUTHOR: Peter Hollemans
     DATE: 2005/07/20
  CHANGES: 2006/10/14, PFH, added getViewVector() method for computing 
             geodetic subpoint 
           2006/12/04, PFH, changed atan to atan2 in getViewVector()
           2007/01/28, PFH, added use of HRPTReaderFactory
           2007/02/09, PFH, moved main functionality to abstract class
           2007/03/05, PFH, modified to use new Spacecraft class

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Imports
// -------
import java.io.*;
import java.util.*;
import terrenus.orbit.*;
import terrenus.telemetry.*;

/**
 * The <code>AVHRRLocator</code> class uses an {@link
 * HRPTDecoder} to compute AVHRR earth locations for a NOAA HRPT
 * data stream.
 */
public class AVHRRLocator
  extends LineScannerLocator {

  // Constants
  // ---------

  /** Standard time unit (STU) period in milliseconds. */
  private static final double STU = 1.001602564;

  /** AVHRR nadir scan time in milliseconds. */
  private static final int NADIR_TIME_OFFSET = (int) (34.2*STU);

  // Variables
  // ---------

  /** The HRPT decoder to use for generating dates for each scan line. */
  private HRPTDecoder decoder;
  
  ////////////////////////////////////////////////////////////

  protected Date getLineDate (
    int line
  ) {

    return (new Date (decoder.getTime (line) + NADIR_TIME_OFFSET));

  } // getLineDate

  ////////////////////////////////////////////////////////////

  /** 
   * Creates a new locator.
   *
   * @param decoder the decoder to use for time and spacecraft
   * information.
   */
  public AVHRRLocator (
    HRPTDecoder decoder
  ) {

    // Initialize
    // ----------
    this.decoder = decoder;
    spacecraft = decoder.getSpacecraft();
    earthModel = EllipsoidModelFactory.getWGS72();
    sensorModel = new SensorModel (earthModel);

    // Get scan step value
    // -------------------
    scanStep = Math.toRadians (Double.parseDouble (spacecraft.getProperty (
      "avhrr.scan.step")));

  } // AVHRRLocator constructor

  ////////////////////////////////////////////////////////////

  /**
   * Performs AVHRR geolocation using the NOAA KLM user's guide
   * Appendix I.2 scan spot method.  The method is designed to
   * produce the same (lat,lon) results as the {@link
   * #getLocations} method but using completely non-modular code.
   * We include it here as a reference implementation to test
   * that the Terrenus code is producing correct results.
   *
   * @param spotLoc the input [line, sample] of the spot.
   * @param earthLoc the output [lat, lon] geodetic earth
   * location of the spot in radians.
   *
   * @see <a href="http://www2.ncdc.noaa.gov/docs/klm/html/i/app-i2.htm">http://www2.ncdc.noaa.gov/docs/klm/html/i/app-i2.htm</a>
   */
  public void getScanSpot (
    int[] spotLoc,
    double[] earthLoc
  ) {

    /**
     * Note that we give comments here in the form KLM I-?? to
     * show where we are in the KLM equations of Appendix I.
     */

    // Setup
    int line = spotLoc[0];
    Date date = getLineDate (line);
    int x = 0, y = 1, z = 2;

    // KLM I-3
    double[] sat = spacecraft.getPosition (date, null);
    double[] satRad = new double[3];
    for (int i = x; i <= z; i++) 
      satRad[i] = sat[i]/(earthModel.getEquator()/1000);
    double[] p = getViewVector (satRad);

    // KLM I-4
    double[] v = spacecraft.getVelocity (date, null);
    double mag = Math.sqrt (v[x]*v[x] + v[y]*v[y] + v[z]*v[z]);
    for (int i = x; i <= z; i++) v[i] /= mag;

    // KLM I-5
    double[] q = {
      v[y]*p[z]-v[z]*p[y],
      v[z]*p[x]-v[x]*p[z],
      v[x]*p[y]-v[y]*p[x]
    };
    mag = Math.sqrt (q[x]*q[x] + q[y]*q[y] + q[z]*q[z]);
    for (int i = x; i <= z; i++) q[i] /= mag;

    // KLM I-9
    int sample = spotLoc[1];
    double sigma = (sample-1023.5)*scanStep;
    double sinSigma = Math.sin (sigma);
    double cosSigma = Math.cos (sigma);
    double[] d = {
      p[x]*cosSigma + q[x]*sinSigma,
      p[y]*cosSigma + q[y]*sinSigma,
      p[z]*cosSigma + q[z]*sinSigma
    };

    // KLM I-19, I-20, I-21
    double re = 6378.135;
    double re2 = re*re;
    double rp = 6356.75052;
    double rp2 = rp*rp;
    double a = d[x]*d[x]/re2 + d[y]*d[y]/re2 + d[z]*d[z]/rp2;
    double b = 2*(sat[x]*d[x]/re2 + sat[y]*d[y]/re2 + sat[z]*d[z]/rp2);
    double c = sat[x]*sat[x]/re2 + sat[y]*sat[y]/re2 + sat[z]*sat[z]/rp2 - 1;

    // KLM I-22
    double r = (-b - Math.sqrt (b*b - 4*a*c)) / (2*a);

    // KLM I-23
    double[] spot = new double[3];
    for (int i = x; i <= z; i++) spot[i] = sat[i] + r*d[i];

    // KLM I-25, I-26, I-27
    double g = spacecraft.getHourAngle (date);
    double cosG = Math.cos (g);
    double sinG = Math.sin (g);
    double[] ecf = {
      spot[x]*cosG + spot[y]*sinG,
      spot[y]*cosG - spot[x]*sinG,
      spot[z]
    };

    // KLM I-29
    earthLoc[1] = Math.atan2 (ecf[y], ecf[x]);

    // KLM I-30
    earthLoc[0] = Math.atan (re2*ecf[z]/(rp2*Math.sqrt (ecf[x]*ecf[x] +
      ecf[y]*ecf[y])));

  } // getScanSpot

  ////////////////////////////////////////////////////////////

  /**
   * Calculates the great circle distance from one location to another.
   * The method uses a calculation for distance on a sphere.<p>
   *
   * Haversine Formula (from R.W. Sinnott, "Virtues of the Haversine",
   * Sky and Telescope, vol. 68, no. 2, 1984, p. 159):<p>
   *   
   * <pre>
   *   dlon = lon2 - lon1
   *   dlat = lat2 - lat1
   *   a = sin^2(dlat/2) + cos(lat1) * cos(lat2) * sin^2(dlon/2)
   *   c = 2 * arcsin(min(1,sqrt(a)))
   *   d = R * c
   * </pre>
   *
   * @param lat1 the first location latitude in radians.
   * @param lon1 the first location longitude in radians.
   * @param lat2 the second location latitude in radians.
   * @param lon2 the second location longitude in radians.
   * 
   * @return the distance between points in kilometres.
   */
  public static double distance (
    double lat1,
    double lon1,
    double lat2,
    double lon2
  ) {

    double dlon = lon2 - lon1;
    double dlat = lat2 - lat1;
    double a = Math.pow (Math.sin (dlat/2), 2) + 
      Math.cos (lat1) * Math.cos (lat2) * Math.pow (Math.sin (dlon/2), 2);
    double c = 2 * Math.asin (Math.min (1, Math.sqrt (a)));
    double d = 6370.997 * c;
    return (d);

  } // distance

  ////////////////////////////////////////////////////////////

  /** Tests this class. */
  public static void main (String[] argv) throws Exception {

    // Create stream
    // -------------
    String name = argv[0];
    TelemetryStream stream = TelemetryStreamFactory.create (name);

    // Compute locations
    // -----------------
    HRPTDecoder decoder = (HRPTDecoder) stream.getDecoder();
    AVHRRLocator locator = new AVHRRLocator (decoder);
    int lines = stream.getFrameCount();
    int[] samples = new int[] {0, 1023, 2047};
    double[] latitude = new double[samples.length];
    double[] longitude = new double[samples.length];
    double[] satZenith = new double[samples.length];
    double[] solZenith = new double[samples.length];
    double[] relAzimuth = new double[samples.length];
    int[] spotLoc = new int[2];
    double[] earthLoc = new double[2];
    for (int i = 0; i < lines; i++) {
      locator.getLocations (i, samples, latitude, longitude, 
        satZenith, solZenith, relAzimuth);
      spotLoc[0] = i;
      System.out.println ("Data at line " + i + ":");
      for (int j = 0; j < samples.length; j++) {
        spotLoc[1] = samples[j];
        locator.getScanSpot (spotLoc, earthLoc);
        double dist = distance (earthLoc[0], earthLoc[1], latitude[j],
          longitude[j]);
        System.out.println ("  Sample [" + samples[j] + "] ");
        System.out.println ("  Scan spot dist = " + dist + " km");
        System.out.println ("  Lat  = " + Math.toDegrees (latitude[j]));
        System.out.println ("  Lon  = " + Math.toDegrees (longitude[j]));
        System.out.println ("  Satz = " + Math.toDegrees (satZenith[j]));
        System.out.println ("  Solz = " + Math.toDegrees (solZenith[j]));
        System.out.println ("  Rela = " + Math.toDegrees (relAzimuth[j]));
      } // for
    } // for

  } // main

  ////////////////////////////////////////////////////////////

} // AVHRRLocator class

////////////////////////////////////////////////////////////////////////
