////////////////////////////////////////////////////////////////////////
/*
     File: AbstractRadiometerDataWriter.java
  PURPOSE: Implements some methods for writing radiometer data.
   AUTHOR: Peter Hollemans
     DATE: 2007/03/13
  CHANGES: 2011/02/05, PFH, added explicit handling of missing data lines

  Copyright (C) 2007-2011, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Imports
// -------
import java.io.*;
import java.util.*;
import terrenus.telemetry.*;
import terrenus.instrument.RadiometerCalibrator.CalibrationType;

/**
 * The <code>AbstractRadiometerDataWriter</code> class implements
 * the main method for the {@link RadiometerDataWriter} class and
 * provides basic writing services.  The child class must
 * implement the various format-specific methods.
 */ 
public abstract class AbstractRadiometerDataWriter 
  implements RadiometerDataWriter {

  // Constants
  // ---------

  /** The angle variable names. */
  private static final String[] ANGLE_NAMES = new String[] {
    "latitude",
    "longitude",
    "sat_zenith",
    "sun_zenith",
    "rel_azimuth"
  };

  // Variables
  // ---------

  /** The origin of the data. */
  protected String origin = "unknown";

  /** The history of the data. */
  protected String history = "unknown";

  ////////////////////////////////////////////////////////////

  /** 
   * Sets the origin string for the data.  The origin is an
   * organization abbreviation that appears on data plots to indicate
   * tha the data was captured and processed by a certain
   * organization.
   *
   * @param origin the data origin.
   */
  public void setOrigin (String origin) { this.origin = origin; }

  ////////////////////////////////////////////////////////////

  /** 
   * Sets the history string for the data.  The history string
   * is a list of the versions and parameters of software used
   * for processing the data.
   *
   * @param history the data history.
   */
  public void setHistory (String history) { this.history = history; }

  ////////////////////////////////////////////////////////////

  /**
   * Creates the specified set of variables in the output.
   *
   * @param varNames the list of variable names to create.
   * @param varUnits the list of units for each variable.
   * @param dims the variable dimensions as [lines, samples].
   *
   * @throws IOException if an error occurred creating the
   * variables.
   */
  protected abstract void createVariables (
    List<String> varNames,
    List<String> varUnits,
    int[] dims
  ) throws IOException;

  ////////////////////////////////////////////////////////////

  /**
   * Writes a line of variable data to the output.
   *
   * @param varIndex the variable index from the {@link
   * #createVariables} name list.
   * @param line the line to write data for.
   * @param lineData the data to write for the line.
   *
   * @throws IOExeption if an error occurred writing the line
   * data.
   */
  protected abstract void writeLineData (
    int varIndex,
    int line,
    double[] lineData
  ) throws IOException;

  ////////////////////////////////////////////////////////////

  /**
   * Writes a line of missing variable data to the output.
   *
   * @param varIndex the variable index from the {@link
   * #createVariables} name list.
   * @param line the line to write data for.
   *
   * @throws IOExeption if an error occurred writing the line
   * data.
   */
  protected abstract void writeMissingLineData (
    int varIndex,
    int line
  ) throws IOException;

  ////////////////////////////////////////////////////////////

  /**
   * Opens a new output sink for radiometer data.
   *
   * @param decoder the stream decoder to read data from.
   * @param radiometer the radiometer instrument to write data for.
   * @param startFrame the starting data frame to write.
   * @param endFrame the ending data frame to write.
   *
   * @throws IOException if an error occurred opening the output
   * sink.
   */  
  protected abstract void openOutput (
    TelemetryDecoder decoder,
    Radiometer radiometer,
    int startFrame,
    int endFrame
  ) throws IOException;

  ////////////////////////////////////////////////////////////

  /**
   * Closes the output sink for radiometer data.
   *
   * @throws IOException if an error occurred closing the output
   * sink.
   */  
  protected abstract void closeOutput () throws IOException;

  ////////////////////////////////////////////////////////////

  public void writeData (
    TelemetryDecoder decoder,
    Radiometer radiometer,
    int startFrame,
    int endFrame,
    CalibrationType visibleCalType,
    CalibrationType thermalCalType
  ) throws IOException {

    // Find usable telemetry frames
    // ----------------------------
    int lines = endFrame - startFrame + 1;
    boolean[] isUsable = new boolean[lines];
    boolean anyUsable = false;
    for (int i = 0; i < lines; i++) {
      isUsable[i] = !decoder.hasErrors (i+startFrame);
      if (isUsable[i]) anyUsable = true;
    } // for
    if (!anyUsable) 
      throw new IOException ("Cannot locate any usable telemetry data");

    // Open output
    // -----------
    openOutput (decoder, radiometer, startFrame, endFrame);

    // Create dimensions
    // -----------------
    int samples = radiometer.getSampleCount();
    int[] dims = new int[] {lines, samples};

    // Create variables
    // ----------------
    List<String> varNames = new ArrayList<String>();
    List<String> varUnits = new ArrayList<String>();
    int channels = radiometer.getChannelCount();
    for (int i = 1; i <= channels; i++) {
      varNames.add (radiometer.getChannelName (i));
      varUnits.add (radiometer.isThermal (i) ? thermalCalType.getUnits() :
        visibleCalType.getUnits());
    } // for
    int angles = ANGLE_NAMES.length;
    for (int i = 0; i < angles; i++) {
      varNames.add (ANGLE_NAMES[i]);
      varUnits.add ("degrees");
    } // for
    createVariables (varNames, varUnits, dims);

    // Loop over each frame
    // --------------------
    int variables = angles + channels;
    double[][] lineData = new double[variables][samples];
    boolean[] writeChannel = new boolean[channels];
    for (int frame = startFrame; frame <= endFrame; frame++) {

      // Read radiometer data
      // --------------------
      int line = frame - startFrame;
      RadiometerData radiometerData = 
        (RadiometerData) decoder.getData (frame, radiometer);
      boolean didWriteLineData = false;
      
      // Handle channel data
      // -------------------
      /**
       * Note: This next if statement has the effect of skipping
       * this line if both it and the previous one were not
       * usable.  But if this line is not usable but the previous
       * line was, the data from the previous line is used for
       * this line.  Thus, singleton missing lines are fixed by
       * duplicating the previous good line.
       */
      boolean thisUsable = isUsable[frame-startFrame];
      boolean lastUsable = (frame != startFrame && 
        isUsable[frame-startFrame-1]);
      if (thisUsable || lastUsable) {

        // Read new channel data
        // ---------------------
        if (thisUsable) {
          for (int channel = 0; channel < channels; channel++) {
            int channelNumber = channel+1;
            CalibrationType calType =
              radiometer.isThermal (channelNumber) ? thermalCalType : visibleCalType;
            try {
              radiometerData.getCalibratedData (channelNumber, calType, 
                lineData[channel]);
              writeChannel[channel] = true;
            } // try
            catch (IOException e) { 
              isUsable[frame-startFrame] = thisUsable = false;
              break;
            } // catch
            catch (IllegalArgumentException e2) {
              writeChannel[channel] = false;
            } // catch
          } // for
        } // if

        // Write channel data
        // ------------------
        if (thisUsable || lastUsable) {
          for (int channel = 0; channel < channels; channel++) {
            if (writeChannel[channel]) 
              writeLineData (channel, line, lineData[channel]);
            else
              writeMissingLineData (channel, line);
          } // for
          didWriteLineData = true;
        } // if

      } // if

      // Write missing data if needed
      // ----------------------------
      if (!didWriteLineData) {
        for (int channel = 0; channel < channels; channel++)
          writeMissingLineData (channel, line);
      } // if

      // Handle angle data
      // -----------------
      int offset = channels;
      radiometerData.getLocationData (lineData[offset], lineData[offset+1],
        lineData[offset+2], lineData[offset+3], lineData[offset+4]);
      for (int angle = offset; angle < variables; angle++) {
        for (int k = 0; k < samples; k++)
          lineData[angle][k] = Math.toDegrees (lineData[angle][k]);
        writeLineData (angle, line, lineData[angle]);
      } // for

    } // for

    // Close output
    // ------------
    closeOutput();

  } // writeData

  ////////////////////////////////////////////////////////////

} // AbstractRadiometerDataWriter class

////////////////////////////////////////////////////////////////////////
