////////////////////////////////////////////////////////////////////////
/*
     File: CWHDFDataWriter.java
  PURPOSE: Writes radiometer data to CWHDF format.
   AUTHOR: Peter Hollemans
     DATE: 2007/03/14
  CHANGES: 2011/02/05, PFH
           - added writeMissingLineData method
           - added check for Double.NaN values when writing data

  Copyright (C) 2007-2011, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Imports
// -------
import java.io.*;
import java.util.*;
import ncsa.hdf.hdflib.*;
import terrenus.util.*;
import terrenus.telemetry.*;

/**
 * The <code>CWHDFDataWriter</code> class writes radiometer data
 * to Hierachical Data Format (HDF) version 4 files.  All
 * available radiometer channels are written in units specified
 * by the {@link #writeData} method.  The metadata written
 * conforms to the CoastWatch HDF version 3.4 specificaion.
 * Latitude, longitude, satellite zenith, solar zenith, and
 * relative azimuth data are also computed and written in
 * degrees.  Latitude and longitude variables are written as HDF
 * 32-bit floats, and all other angle data and radiometer
 * channels as HDF 16-bit signed integers scaled by 0.01.
 */
public class CWHDFDataWriter 
  extends AbstractRadiometerDataWriter {

  // Constants
  // ---------

  /** The number of milliseconds in one day. */
  private static final int MSEC_PER_DAY = 86400000;

  /** The set of variable names to write as float data. */
  private static final HashSet<String> FLOAT_VARS = 
    new HashSet<String> (Arrays.asList (new String[] {
      "latitude", 
      "longitude"
    }));

  /** The data value stride for writing each data line. */
  private static final int[] STRIDE = new int[] {1, 1};

  // Variables
  // ---------

  /** The output file name. */
  private String output;

  /** The HDF SDID for file access. */
  private int sdid;
  
  /** The variable SDID index array. */
  private int[] varRefs;

  /** The array of flags that indicate whether data should be scaled. */
  private boolean[] isScaled;

  /** The array of temporary floating point values for writing. */
  private float[] floatData;

  /** The array of temporary short values for writing. */
  private short[] shortData;

  /** The data value count for writing each data line. */
  private int[] count;

  /** The data value start for writing each data line. */
  private int[] start;

  ////////////////////////////////////////////////////////////

  /** Creates a new writer. */
  public CWHDFDataWriter (
    String output
  ) {

    this.output = output;

  } // CWHDFDataWriter constuctor

  ////////////////////////////////////////////////////////////

  protected void createVariables (
    List<String> varNames,
    List<String> varUnits,
    int[] dims
  ) throws IOException {

    // Create variable reference array
    // -------------------------------
    int variables = varNames.size();
    varRefs = new int[variables];
    isScaled = new boolean[variables];
    for (int i = 0; i < variables; i++) {
      String varName = varNames.get (i);
      String varUnit = varUnits.get (i);
      isScaled[i] = !FLOAT_VARS.contains (varName);
      int hdfType = (isScaled[i] ? HDFConstants.DFNT_INT16 : 
        HDFConstants.DFNT_FLOAT32);
      varRefs[i] = HDFServices.createVariable (sdid, varName, hdfType, 
        varUnit, dims, false);
    } // for

  } // createVariables

  ////////////////////////////////////////////////////////////

  protected void writeMissingLineData (
    int varIndex,
    int line
  ) throws IOException {
  
    /**
     * We do nothing here because HDF fills in the missing values
     * for us when the data is not written.  This method may need
     * an implementation in other data formats, for example NetCDF.
     */
  
  } // writeMissingLineData

  ////////////////////////////////////////////////////////////

  protected void writeLineData (
    int varIndex,
    int line,
    double[] lineData
  ) throws IOException {

    // Convert data to output values
    // -----------------------------
    Object data;
    if (isScaled[varIndex]) {
      for (int i = 0; i < lineData.length; i++) {
        if (Double.isNaN (lineData[i]))
          shortData[i] = Short.MIN_VALUE;
        else
          shortData[i] = (short) Math.round (lineData[i] * 100);
      } // for
      data = shortData;
    } // if
    else {
      for (int i = 0; i < lineData.length; i++) {
        if (Double.isNaN (lineData[i]))
          floatData[i] = Float.NaN;
        else
          floatData[i] = (float) lineData[i];
      } // for
      data = floatData;
    } // else

    // Write data
    // ----------
    start[0] = line;
    HDFServices.writeVariable (sdid, varRefs[varIndex], false, start, 
      STRIDE, count, data);

  } // writeLineData

  ////////////////////////////////////////////////////////////

  protected void openOutput (
    TelemetryDecoder decoder,
    Radiometer radiometer,
    int startFrame,
    int endFrame
  ) throws IOException {

    // Create file
    // -----------
    try {
      sdid = HDFLibrary.SDstart (output, HDFConstants.DFACC_CREATE);
      if (sdid < 0) throw new HDFException();
    } // try
    catch (HDFException e) {
      throw new IOException ("Error creating HDF file");
    } // catch

    // Write global attributes
    // -----------------------
    HDFServices.writeAtt (sdid, "satellite", 
      decoder.getSpacecraft().getName());
    HDFServices.writeAtt (sdid, "sensor", radiometer.getName());
    HDFServices.writeAtt (sdid, "origin", origin);
    HDFServices.writeAtt (sdid, "cwhdf_version", "3.4");
    HDFServices.writeAtt (sdid, "pass_type", "unknown");
    long msec = decoder.getTime (startFrame);
    HDFServices.writeAtt (sdid, "pass_date", 
      new int[] {(int) (msec/MSEC_PER_DAY)});
    HDFServices.writeAtt (sdid, "start_time", 
      new double[] {(msec%MSEC_PER_DAY)/1000.0});
    HDFServices.writeAtt (sdid, "projection_type", "swath");
    int lines = endFrame - startFrame + 1;
    HDFServices.writeAtt (sdid, "rows", new int[] {lines});
    int samples = radiometer.getSampleCount();
    HDFServices.writeAtt (sdid, "cols", new int[] {samples});
    HDFServices.writeAtt (sdid, "history", history);

    // Create various data arrays
    // --------------------------
    floatData = new float[samples];
    shortData = new short[samples];
    count = new int[] {1, samples};
    start = new int[] {0, 0};

  } // openOutput

  ////////////////////////////////////////////////////////////

  protected void closeOutput () throws IOException {

    try {
      if (!HDFLibrary.SDend (sdid)) throw new HDFException();
    } // try
    catch (HDFException e) {
      throw new IOException ("Error closing HDF file");
    } // catch

  } // closeOutput

  ////////////////////////////////////////////////////////////

} // CWHDFDataWriter class

////////////////////////////////////////////////////////////////////////
