////////////////////////////////////////////////////////////////////////
/*
     FILE: MVISRData.java
  PURPOSE: Interface for data from radiometers.
   AUTHOR: Peter Hollemans
     DATE: 2007/03/16
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Import 
// ------
import java.io.*;
import java.nio.*;
import terrenus.telemetry.*;
import terrenus.instrument.RadiometerCalibrator.Record;
import terrenus.instrument.RadiometerCalibrator.CalibrationType;

/**
 * The <code>MVISRData</code> class holds data from an MVISR instrument
 * on the FY1 series spacecrafts. 
 *
 * @see MVISR
 * @see terrenus.orbit.FY1Spacecraft
 */
public class MVISRData
  implements RadiometerData {

  // Variables
  // ---------

  /** The frame number for this data. */
  private int frame;

  /** The decoder to use for frame data. */
  private CHRPTDecoder decoder;

  /** The calibrator to use for calibration. */
  private MVISRCalibrator calibrator;

  /** The calibration record for this data. */
  private Record record;

  /** The location generator. */
  private MVISRLocator locator;

  /** The temporary count data array. */
  private int[] tempCountData = new int[MVISR.SAMPLES];

  /** The sample values array. */
  private static int[] sampleValues;

  ////////////////////////////////////////////////////////////

  /**
   * Creates a new MVISR data object for the specified telemetry frame.
   *
   * @param frame the frame number for the data.
   * @param decoder the CHRPT decoder to use for data.
   * @param calibrator the calibrator to use for calibration.
   * @param record the calibration record for this data or null
   * if one cannot be created.
   * @param locator the location generator.
   */
  public MVISRData (
    int frame,
    CHRPTDecoder decoder,
    MVISRCalibrator calibrator,
    Record record,
    MVISRLocator locator
  ) {

    this.frame = frame;
    this.decoder = decoder;
    this.calibrator = calibrator;
    this.record = record;
    this.locator = locator;

  } // MVISRData constructor

  ////////////////////////////////////////////////////////////

  public Instrument getInstrument() { return (MVISR.getInstance()); }

  ////////////////////////////////////////////////////////////

  public int[] getCountData (
    int channel, 
    int[] countData
  ) throws IOException {

    // Create count data array
    // -----------------------
    if (countData == null) countData = new int[MVISR.SAMPLES];

    // Get count data for channel
    // --------------------------
    ShortBuffer earthData = decoder.getStream().getBlock (frame, 
      CHRPTStream.EARTH_BLOCK).asShortBuffer();
    int bufOffset = channel - 1;
    for (int i = 0; i < MVISR.SAMPLES; i++)
      countData[i] = earthData.get (i*10 + bufOffset);

    return (countData);

  } // getCountData

  ////////////////////////////////////////////////////////////

  public double[] getCalibratedData (
    int channel,
    CalibrationType calType,
    double[] calData
  ) throws IOException {

    // Check calibration record
    // ------------------------
    if (record == null)
      throw new IOException ("No calibration record available");

    // Create calibrated data array
    // ----------------------------
    if (calData == null) calData = new double[MVISR.SAMPLES];

    // Get count data and calibrate
    // ----------------------------
    getCountData (channel, tempCountData);
    calibrator.calibrate (tempCountData, calData, channel, calType, record);

    return (calData);

  } // getCalibratedData

  ////////////////////////////////////////////////////////////

  public void getLocationData (
    double[] latitude,
    double[] longitude,
    double[] satZenith,
    double[] solZenith,
    double[] relAzimuth
  ) {

    // Create sample values array
    // --------------------------
    if (sampleValues == null) {
      sampleValues = new int[MVISR.SAMPLES];
      for (int i = 0; i < MVISR.SAMPLES; i++) sampleValues[i] = i;
    } // if

    // Compute location and angle data
    // -------------------------------
    locator.getLocations (frame, sampleValues, latitude, longitude,
      satZenith, solZenith, relAzimuth);

  } // getLocationData

  ////////////////////////////////////////////////////////////

  /** Tests this class. */
  public static void main (String[] argv) throws Exception {

    int i = 0;
    String name = argv[i++];
    TelemetryStream stream = TelemetryStreamFactory.create (name);
    TelemetryDecoder decoder = stream.getDecoder();

    int startFrame = Integer.parseInt (argv[i++]);
    int endFrame = Integer.parseInt (argv[i++]);
    int startSample = Integer.parseInt (argv[i++]);
    int endSample = Integer.parseInt (argv[i++]);
    MVISR mvisr = MVISR.getInstance();
    int[][] countData = new int[10][2048];
    double[] latData = new double[2048];
    double[] lonData = new double[2048];
    double[] junk = new double[2048];
    for (int frame = startFrame; frame <= endFrame; frame++) {
      MVISRData data = (MVISRData) decoder.getData (frame, mvisr);
      for (int channel = 0; channel < 10; channel++) 
        data.getCountData (channel+1, countData[channel]);
      data.getLocationData (latData, lonData, junk, junk, junk);
      for (int sample = startSample; sample <= endSample; sample++) {
        System.out.print (frame + " " + sample + " ");
        System.out.printf ("%.2f ", Math.toDegrees (latData[sample]));
        System.out.printf ("%.2f ", Math.toDegrees (lonData[sample]));
        for (int channel = 0; channel < 10; channel++) 
          System.out.print (countData[channel][sample] + " ");
        System.out.println();
      } // for
    } // for

  } // main

  ////////////////////////////////////////////////////////////

} // MVISRData class

////////////////////////////////////////////////////////////////////////
