////////////////////////////////////////////////////////////////////////
/*
     FILE: RadiometerCalibrator.java
  PURPOSE: Interface for calibrating radiometer data.
   AUTHOR: Peter Hollemans
     DATE: 2007/03/03
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Imports
// -------
import java.io.*;

/**
 * The <code>RadiometerCalibrator</code> interface is for
 * calibrating raw count data from multichannel radiometers.  A
 * calibration {@link Record} holds calibration coefficients
 * computed using data from a number of scan lines.  The {@link
 * #getCalibration(int)}, {@link #getCalibration(int,int)}, and
 * {@link #calibrate} methods are low level developer methods for
 * calibrating data -- users would normally go through the {@link
 * RadiometerData} interface for calibrating data.
 */
public interface RadiometerCalibrator {

  // Constants
  // ---------

  /** The types of calibration. */
  public enum CalibrationType {
    COUNTS, REFLECTANCE, ALBEDO, RADIANCE, CELSIUS, KELVIN;
    public String getUnits () {
      switch (this) {
      case COUNTS: return ("counts");
      case REFLECTANCE: return ("1");
      case ALBEDO: return ("percent");
      case RADIANCE: return ("mW/(m^2 sr cm^-1)");
      case CELSIUS: return ("celsius");
      case KELVIN: return ("kelvin");
      default: return ("unknown");
      } // switch
    } // getUnits
  };
  
  ////////////////////////////////////////////////////////////

  /**
   * The <code>Record</code> class holds radiometer calibration
   * coefficients for a sequence of scan lines.
   */
  public class Record {

    // Variables
    // ---------

    /** The calibration coefficients as [channel][coefficient]. */
    private double[][] coefs;

    /** The scan line boundaries used for this calibration record. */
    private int startLine, endLine;

    ////////////////////////////////////////////////////////

    /** 
     * Creates a new calibration record. 
     *
     * @param coefs the calibration coefficients as [channel][coef].
     * @param startLine the starting calibration line.
     * @param endLine the ending calibration line.
     */
    public Record (
      double[][] coefs,
      int startLine,
      int endLine
    ) {

      this.coefs = coefs;
      this.startLine = startLine;
      this.endLine = endLine;

    } // Record constructor

    ////////////////////////////////////////////////////////

    /** 
     * Gets the coefficients for the specified channel. 
     *
     * @param channel the channel to calibrate in the range [1..CHANNELS].
     */
    public double[] getCoefs (int channel) { return (coefs[channel-1]); }

    ////////////////////////////////////////////////////////

    /** Gets the starting calibration line. */
    public int getStartLine () { return (startLine); }

    ////////////////////////////////////////////////////////

    /** Gets the ending calibration line. */
    public int getEndLine () { return (endLine); }

    ////////////////////////////////////////////////////////
    
    /** Converts this record to a string. */
    public String toString () {

      StringBuffer buffer = new StringBuffer();
      for (int channel = 0; channel < coefs.length; channel++) {
        buffer.append ("channel" + (channel+1) + "=[");
        for (int i = 0; i < coefs[channel].length; i++) {
          buffer.append (coefs[channel][i]);
          if (i < coefs[channel].length - 1) buffer.append (",");
        } // for
        buffer.append ("]\n");
      } // for
      buffer.append ("startLine=" + startLine + ",");
      buffer.append ("endLine=" + endLine);

      return (buffer.toString());

    } // toString

    ////////////////////////////////////////////////////////

  } // Record

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the calibration record for the specified scan line.
   * Generally calibration requires a number of radiometer scan
   * lines to compute.  The returned record represents the best
   * available calibration data based on lines surrounding or
   * near the specified scan line.
   *
   * @param line the line to retrieve calibration for.
   *
   * @return the calibration record or null if no record is
   * available.
   *
   * @throws IOException if a problem occurred accessing the data.
   */
  public Record getCalibration (
    int line
  ) throws IOException;

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the calibration record for the specified scan line
   * range.  Only scan lines in the specified range are used to
   * compute the calibration record.
   *
   * @param startLine the starting scan line for calibration.
   * @param endLine the ending scan line for calibration.
   *
   * @return the calibration record or null if one could not be
   * computed.
   *
   * @throws IOException if a problem occurred accessing the data.
   */
  public Record getCalibration (
    int startLine,
    int endLine
  ) throws IOException;

  ////////////////////////////////////////////////////////////

  /**
   * Calibrates radiometer data from raw count data to scientific
   * units.
   *
   * @param countData the raw radiometer count data.
   * @param calData the output calibrated data.
   * @param channel the channel to calibrate in the range [1..CHANNELS].
   * @param calType the calibration type to apply.
   * @param record the calibration record to use.
   */
  public void calibrate (
    int[] countData,
    double[] calData,
    int channel,
    CalibrationType calType,
    Record record
  );

  ////////////////////////////////////////////////////////////

} // RadiometerCalibrator interface

////////////////////////////////////////////////////////////////////////
