////////////////////////////////////////////////////////////////////////
/*
     FILE: RadiometerData.java
  PURPOSE: Interface for data from radiometers.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/23
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Imports
// -------
import java.io.*;

/**
 * The <code>RadiometerData</code> interface is for multichannel
 * radiometer data from a spacecraft.
 *
 * @see Radiometer
 */
public interface RadiometerData
  extends InstrumentData {

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the raw count data for the radiometer channel.
   *
   * @param channel the channel to retrieve data for.
   * @param countData the destination data array or null to create.
   *
   * @return the count data array.
   *
   * @throws IOException if an error occurred reading the data.
   * @throws IllegalArgumentException if the raw count data for the
   * specified channel is not available (for example, the channel may
   * be turned off).
   */
  public int[] getCountData (
    int channel, 
    int[] countData
  ) throws IOException;

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the calibrated data for the radiometer channel.
   *
   * @param channel the channel to retrieve data for in the range
   * [1..N].
   * @param calType the calibration type.
   * @param calData the destination data array or null to create.
   *
   * @return the calibrated data array.
   *
   * @throws IOException if an error occurred reading the data or
   * no calibration data was available.
   * @throws IllegalArgumentException if the raw count data for the
   * specified channel is not available (for example, the channel may
   * be turned off).
   */
  public double[] getCalibratedData (
    int channel,
    RadiometerCalibrator.CalibrationType calType,
    double[] calData
  ) throws IOException;

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the earth locations and angles for the radiometer data.
   * Locations are geodetic latitude and longitude.  All angles are in
   * radians.
   *
   * @param latitude the output latitude array.
   * @param longitude the output longitude array.
   * @param satZenith the output satellite zenith array.
   * @param solZenith the output solar zenith array.
   * @param relAzimuth the output relative azimuth array.
   */
  public void getLocationData (
    double[] latitude,
    double[] longitude,
    double[] satZenith,
    double[] solZenith,
    double[] relAzimuth
  );

  ////////////////////////////////////////////////////////////

} // RadiometerData interface

////////////////////////////////////////////////////////////////////////
