////////////////////////////////////////////////////////////////////////
/*
     FILE: AbstractSpacecraft.java
  PURPOSE: Retrieves spacecraft information.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/20
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

// Imports
// -------
import java.util.*;
import java.io.*;
import java.text.*;
import java.net.*;
import terrenus.instrument.*;

/**
 * The <code>AbstractSpacecraft</code> class implements some of
 * the main methods of the {@link Spacecraft} interface to
 * deliver position, velocity, hour angle, name, and launch date.
 * Subclasses must implement the {@link
 * Spacecraft#getInstruments} method.
 */
public abstract class AbstractSpacecraft 
  implements Spacecraft {

  // Constants
  // ---------

  /** The name property. */
  private static final String NAME_PROP = "long.name";

  /** The launch date property. */
  private static final String LAUNCH_PROP = "launch.date";

  // Variables
  // ---------

  /** The locator to determine orbital model information. */
  private SpacecraftLocator locator;

  /** The spacecraft properties. */
  private Properties props;

  /** The spacecraft NORAD orbital element ID. */
  private int noradId;
  
  /** The date used to initialize the locator computations. */
  private Date locatorDate;

  /** The spacecraft name. */
  private String name;

  /** The spacecraft launch date. */
  private Date launchDate;

  /** The map of NORAD ID to properties file. */
  private static Map<Integer,String> propsFileMap;

  /** The parser for launch date. */
  private static SimpleDateFormat dateParser;

  ////////////////////////////////////////////////////////////

  static {

    // Create map of NORAD IDs to files
    // --------------------------------
    propsFileMap = new HashMap<Integer,String>();
    Properties fileProps = getProperties ("norad.properties");
    for (Object property : fileProps.keySet()) 
      propsFileMap.put (Integer.parseInt ((String) property), 
        fileProps.getProperty ((String) property));

    // Create launch date parser
    // -------------------------
    dateParser = new SimpleDateFormat ("yyyy/MM/dd");
    dateParser.setTimeZone (TimeZone.getTimeZone ("GMT+0"));

  } // static

  ////////////////////////////////////////////////////////////

  /** Gets the specified properties file as a properties object. */
  private static Properties getProperties (
    String propsFile
  ) {

    Properties props;
    try {
      InputStream stream = 
        AbstractSpacecraft.class.getResourceAsStream (propsFile);
      if (stream == null) throw new IOException();
      props = new Properties();
      props.load (stream);
      stream.close();
    } // try 
    catch (IOException e) {
      throw new RuntimeException ("Error loading properties file '" + 
        propsFile + "'");
    } // catch

    return (props);

  } // getProperties

  ////////////////////////////////////////////////////////////

  /**
   * Creates a new spacecraft using the NORAD ID.
   *
   * @param noradId the NORAD orbital element ID.
   *
   * @throws RuntimeException if the spacecraft properties cannot be
   * found or parsed.
   */
  public AbstractSpacecraft (
    int noradId
  ) {

    // Initialize
    // ----------
    this.noradId = noradId;
    String propsFile = propsFileMap.get (noradId);
    if (propsFile == null) 
      throw new RuntimeException ("Cannot locate spacecraft properties file");
    props = getProperties (propsFile);

    // Get name and launch date
    // ------------------------
    name = props.getProperty (NAME_PROP);
    try { 
      launchDate = dateParser.parse (props.getProperty (LAUNCH_PROP)); 
    } // try
    catch (ParseException e) { 
      throw new RuntimeException ("Error parsing launch date");
    } // catch

  } // AbstractSpacecraft constructor

  ////////////////////////////////////////////////////////////

  public String getName () { return (name); }

  ////////////////////////////////////////////////////////////

  /**
   * Updates the spacecraft locator for the specified date.  If
   * the locator doesn't exist yet, it is created.  If the
   * locator is set for a different date, it is initialized with
   * the new date.
   */
  private void updateLocator (
    Date date
  ) {

    // Check existing locator
    // ----------------------
    if (locator != null) {
      if (locatorDate.equals (date)) return;
      else if (!locator.isWithinTolerance (date)) { locator = null; }
    } // if

    // Create new locator
    // ------------------
    if (locator == null) {
      locator = new SpacecraftLocator (noradId, date);
    } // if

    // Set locator date
    // ----------------
    locator.setDate (date);
    locatorDate = date;

  } // updateLocator

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the Greenwich hour angle.
   *
   * @param date the date for Greenwich hour angle retrieval.
   *
   * @return the Greenwich hour angle in radians.
   */
  public double getHourAngle (
    Date date 
  ) {

    updateLocator (date);
    return (locator.getHourAngle());

  } // getHourAngle

  ////////////////////////////////////////////////////////////

  /**
   * Gets the position of the spacecraft in earth-centered
   * inertial coordinates.
   * 
   * @param date the date for position retrieval.
   * @param output the output position array to fill.  If null, an
   * output array is created.
   *
   * @return the output position array as [x,y,z] in kilometers.
   */
  public double[] getPosition (
    Date date,
    double[] output
  ) {

    updateLocator (date);
    return (locator.getPosition (output));

  } // getPosition

  ////////////////////////////////////////////////////////////

  /**
   * Gets the velocity of the spacecraft in earth-centered
   * inertial coordinates.
   * 
   * @param date the date for velocity retrieval.
   * @param output the output velocity array to fill.  If null, an
   * output array is created.
   *
   * @return the output velocity array as [x,y,z] in
   * kilometers/second.
   */
  public double[] getVelocity (
    Date date,
    double[] output
  ) {

    updateLocator (date);
    return (locator.getVelocity (output));

  } // getVelocity

  ////////////////////////////////////////////////////////////

  public Date getLaunchDate() { return (launchDate); }

  ////////////////////////////////////////////////////////////

  public String getProperty (String name) { 

    return (props.getProperty (name)); 

  } // getProperty

  ////////////////////////////////////////////////////////////

  /** Tests this class. */
  public static void main (String[] argv) throws Exception {

    // Add current elements to manager
    // -------------------------------
    OrbitalElementManager manager = OrbitalElementManager.getInstance();
    manager.registerSource (new CurrentElementSource (
      new URL ("http://www.celestrak.net/NORAD/elements/weather.txt")));

    // Create a spacecraft
    // -------------------
    Spacecraft craft = new AbstractSpacecraft (Integer.parseInt (argv[0])) {
        public List<Instrument> getInstruments () { return (null); }
      };

    // Create date formatter
    // ---------------------
    SimpleDateFormat dateFormat = 
      new SimpleDateFormat ("yyyy/MM/dd HH:mm:ss Z");
    dateFormat.setTimeZone (TimeZone.getTimeZone ("GMT+0"));

    // Print out the basic info
    // ------------------------
    System.out.println ("Name:           " + craft.getName());
    System.out.println ("Launch date:    " + 
      dateFormat.format (craft.getLaunchDate()));

    // Compute the position
    // --------------------
    Date date = new Date();
    System.out.println ("Current date:   " + dateFormat.format (date));
    double[] posECI = craft.getPosition (date, null);
    EllipsoidModel model = EllipsoidModelFactory.getWGS72();
    double g = craft.getHourAngle (date);
    double cosG = Math.cos (g);
    double sinG = Math.sin (g);
    double[] pos = {
      posECI[0]*cosG + posECI[1]*sinG,
      posECI[1]*cosG - posECI[0]*sinG,
      posECI[2]
    };
    for (int i = 0; i < 3; i++) pos[i] /= (model.getEquator()/1000);
    double[] geoPos = model.ECEFToGD (pos, 0, null);
    System.out.printf ("Position (ECF): %.2f, %.2f, %.2f (norm)\n", pos[0], 
      pos[1], pos[2]);
    System.out.printf ("Position (GEO): %.2f N, %.2f E (deg)\n", 
      Math.toDegrees (geoPos[0]), Math.toDegrees (geoPos[1]));

    // Compute the velocity
    // --------------------
    double[] velECI = craft.getVelocity (date, null);
    double[] vel = {
      velECI[0]*cosG + velECI[1]*sinG,
      velECI[1]*cosG - velECI[0]*sinG,
      velECI[2]
    };
    System.out.printf ("Velocity (ECF): %.2f, %.2f, %.2f (km/s)\n", vel[0],
      vel[1], vel[2]);

    // Compute sun angle
    // -----------------
    Sun.setDate (date);
    double[] sunECI = Sun.getPosition (null);
    double[] sun = {
      sunECI[0]*cosG + sunECI[1]*sinG,
      sunECI[1]*cosG - sunECI[0]*sinG,
      sunECI[2]
    };
    double[] normal = model.getNormal (pos, null);
    double[] sunDir = VectorOp.subtract (sun, pos, null);
    VectorOp.normalize (sunDir, sunDir);
    double solZenith = Math.acos (VectorOp.dot (normal, sunDir));
    System.out.printf ("Solar zenith:   %.2f (deg)\n", 
      Math.toDegrees (solZenith));

  } // main

  ////////////////////////////////////////////////////////////

} // AbstractSpacecraft class

////////////////////////////////////////////////////////////////////////
