////////////////////////////////////////////////////////////////////////
/*
     FILE: ArchiveElementSource.java
  PURPOSE: Provides two-line orbital elements from archived data.
   AUTHOR: Peter Hollemans
     DATE: 2005/10/19
  CHANGES: n/a

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

// Imports
// -------
import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;

/**
 * The <code>ArchiveElementSource</code> class provides access to
 * two-line elements in an archive-style format.  The data contains
 * elements from various different dates in the "historical archives"
 * format provided by the CelesTrak website:
 * <pre>
 *   ELEMENT LINE 1
 *   ELEMENT LINE 2
 *   ELEMENT LINE 1
 *   ELEMENT LINE 2
 *   ...
 * </pre>
 * Elements are assumed to be all from the same satellite.  The
 * two-line element set with the closest matching date is used by
 * {@link #getElements}.
 */
public class ArchiveElementSource 
  extends AbstractOrbitalElementSource {

  // Variables
  // ---------

  /** The URL to use for the element source. */
  private URL url;

  ////////////////////////////////////////////////////////////

  /** 
   * Creates a new local source.
   *
   * @param url the URL to use for element data.
   */
  public ArchiveElementSource (
    URL url
  ) {
    
    this.url = url;

  } // ArchiveElementSource constructor

  ////////////////////////////////////////////////////////////

  public String[] getElements (
    int satid,
    Date date,
    double maxDays
  ) {

    String[] closestElements = new String[2];
    double closestDays = -1;
    try {

      // Open the URL
      // ------------
      BufferedReader reader = new BufferedReader (
        new InputStreamReader (url.openStream()));

      // Search for the closest element date
      // -----------------------------------
      String[] elements = new String[] {reader.readLine(), reader.readLine()};
      boolean eof = (elements[0] == null || elements[1] == null);
      while (!eof) {
        
        // Get satellite id
        // ----------------
        int id = -1;
        try { id = Integer.parseInt (elements[0].substring (2, 7)); }
        catch (NumberFormatException e) { }
        /**
         * We assume here that if the satellite ID we're looking for does not
         * match the IDs in this file, we're looking in the wrong file.
         */
        if (satid != id) break;

        // Get epoch offset
        // ----------------
        double days = Math.abs (getEpochOffset (elements, date));
        if (closestDays < 0 || days < closestDays) {
          closestElements[0] = elements[0]; 
          closestElements[1] = elements[1]; 
          closestDays = days;
        } // if

        // Read next elements
        // ------------------
        elements[0] = reader.readLine(); 
        elements[1] = reader.readLine();
        eof = (elements[0] == null || elements[1] == null);

      } // while
      reader.close();
      
    } // try
    catch (IOException e) { return (null); }

    // Check minimum offset
    // --------------------
    if (closestDays < 0 || closestDays > maxDays) 
      return (null);

    return (closestElements);

  } // getElements

  ////////////////////////////////////////////////////////////

  /** Tests this class. */
  public static void main (String[] argv) throws Exception {

    URL url = new URL (argv[0]);
    int satid = Integer.parseInt (argv[1]);
    long offset = Long.parseLong (argv[2]);
    ArchiveElementSource source = new ArchiveElementSource (url);
    Date then = new Date (new Date().getTime() + offset);
    String[] elements = source.getElements (satid, then, 5);
    if (elements == null) 
      System.out.println ("Cannot find elements for " + satid);
    else {
      DecimalFormat format = new DecimalFormat ("0.00");
      System.out.println ("Found elements with age " + 
        format.format (source.getEpochOffset (elements, then)) + " days:");
      System.out.println (elements[0]);
      System.out.println (elements[1]);
    } // else

  } // main

  ////////////////////////////////////////////////////////////

} // ArchiveElementSource class

////////////////////////////////////////////////////////////////////////
