////////////////////////////////////////////////////////////////////////
/*
     FILE: CurrentElementSource.java
  PURPOSE: Provides two-line orbital elements from current data.
   AUTHOR: Peter Hollemans
     DATE: 2005/10/19
  CHANGES: n/a

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

// Imports
// -------
import java.io.*;
import java.util.*;
import java.net.*;
import java.text.*;

/**
 * The <code>CurrentElementSource</code> class provides access to
 * two-line elements in a current-style format.  The data contains
 * elements from various different satellites in the "current data"
 * format provided by the CelesTrak website:
 * <pre>
 *   SATELLITE NAME #1
 *   ELEMENT LINE 1
 *   ELEMENT LINE 2
 *   SATELLITE NAME #2
 *   ELEMENT LINE 1
 *   ELEMENT LINE 2
 *   ...
 * </pre>
 * Satellite names are ignored.  Any line starting with the string
 * "1&lt;space&gt;" is considered to be the first line of a two-line
 * element set.  Only the first two-line element set matching the
 * requested satellite is used by {@link #getElements}.
 */
public class CurrentElementSource 
  extends AbstractOrbitalElementSource {

  // Variables
  // ---------

  /** The URL to use for the element source. */
  private URL url;

  ////////////////////////////////////////////////////////////

  /** 
   * Creates a new network source.
   *
   * @param url the URL to use for element data.
   */
  public CurrentElementSource (
    URL url
  ) {
    
    this.url = url;

  } // CurrentElementSource constructor

  ////////////////////////////////////////////////////////////

  public String[] getElements (
    int satid,
    Date date,
    double maxDays
  ) {

    String[] elements;
    try {

      // Open the URL
      // ------------
      BufferedReader reader = new BufferedReader (
        new InputStreamReader (url.openStream()));

      // Find the satellite id
      // ---------------------
      String line;
      while ((line = reader.readLine()) != null) {
        if (line.startsWith ("1 ")) {
          int id = -1;
          try { id = Integer.parseInt (line.substring (2, 7)); }
          catch (NumberFormatException e) { }
          if (satid == id) break;
        } // if
      } // while
      if (line == null) return (null);

      // Get two line element data
      // -------------------------
      elements = new String[] {line, reader.readLine()};
      reader.close();

    } // try
    catch (IOException e) { return (null); }

    // Check data for time difference
    // ------------------------------
    if (Math.abs (getEpochOffset (elements, date)) > maxDays) 
      return (null);

    return (elements);

  } // getElements

  ////////////////////////////////////////////////////////////

  /** Tests this class. */
  public static void main (String[] argv) throws Exception {

    URL url = new URL (argv[0]);
    int satid = Integer.parseInt (argv[1]);
    CurrentElementSource source = new CurrentElementSource (url);
    Date now = new Date();
    String[] elements = source.getElements (satid, now, 5);
    if (elements == null) 
      System.out.println ("Cannot find elements for " + satid);
    else {
      DecimalFormat format = new DecimalFormat ("0.00");
      System.out.println ("Found elements with age " + 
        format.format (source.getEpochOffset (elements, now)) + " days:");
      System.out.println (elements[0]);
      System.out.println (elements[1]);
    } // else

  } // main

  ////////////////////////////////////////////////////////////

} // CurrentElementSource class

////////////////////////////////////////////////////////////////////////
