////////////////////////////////////////////////////////////////////////
/*
     FILE: EllipsoidModel.java
  PURPOSE: Performs coordinate transformations on an ellipsoid.
   AUTHOR: Peter Hollemans
     DATE: 2005/09/10
  CHANGES: 2006/10/13, PFH, fixed problem with ellipsoid A matrix
           2006/10/14, PFH, added GCToGDLat2 method
           2006/12/04, PFH, added convergence test in GC to GD lat loops

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

// Imports
// -------
import java.util.*;

/**
 * The <code>EllipsoidModel</code> class converts between various
 * different coordinate systems on an ellipsoid.  In reality, the
 * ellipsoid is just an oblate spheroid because it can have a polar
 * radius that is different from its equatorial radius, but no third
 * radius value.  This describes the Earth and other planets fairly
 * well.<p>
 *
 * Many of the methods in this class deal with cartesian (x,y,z)
 * coordinates and distances in radian units, that is, all
 * distances are divided by the equatorial radius of the
 * ellipsoid (so a point on the equator is a distance of 1 from
 * the center).  This makes the form of the equations simpler in
 * some cases.
 */
public class EllipsoidModel {

  // Constants
  // ---------

  /** The index of the x coordinate. */
  private static final int X = 0;

  /** The index of the y coordinate. */
  private static final int Y = 1;

  /** The index of the z coordinate. */
  private static final int Z = 2;

  /** The ellipsoid equation coefficient c. */
  private static final double ELL_C = -1;

  /** The index of the latitude coordinate. */
  private static final int LAT = 0;

  /** The index of the longitude coordinate. */
  private static final int LON = 1;

  /** The GC to GD latitude maximum loop iterations. */
  private static final int MAX_ITERATIONS = 100;

  // Variables
  // ---------

  /** The equatorial radius in meters. */
  private double rem;

  /** The equatorial radius in equator radius units. */
  private double re;

  /** The equatorial radius squared. */
  private double re2;

  /** The polar radius in meters. */
  private double rpm;

  /** The polar radius in equator radius units. */
  private double rp;

  /** The polar radius squared. */
  private double rp2;

  /** The equatorial radius over polar radius, all squared. */
  private double re_o_rp_2;

  /** The equatorial radius over polar radius, all to the fourth. */
  private double re_o_rp_4;

  /** The ellipsoid equation (x.Ax + b.x + c) matrix A. */
  private double[][] ell_A;

  /** Temporary vector for intersection computations. */
  private double[] tempVector = new double[3];

  /** Temporary quadratic solution array for intersection computations. */
  private double[] solution = new double[2];

  /** The quadratic equation (alpha*t^2 + beta*t + gamma = 0) gamma value. */
  private double gamma;

  /** The last position vector for intersection computations. */
  private double[] lastPosVector = new double[3];

  ////////////////////////////////////////////////////////////

  /**
   * Creates a new ellipsoid model.
   *
   * @param rEquator the radius of the equator in meters.
   * @param rPole the radius of the pole in meters.
   */
  public EllipsoidModel (
    double rEquator,
    double rPole
  ) {

    // Initialize
    // ----------
    rem = rEquator;
    re = 1;
    re2 = 1;
    rpm = rPole;
    rp = rpm/rem;
    rp2 = rp*rp;
    re_o_rp_2 = Math.pow (rem/rpm, 2);
    re_o_rp_4 = Math.pow (rem/rpm, 4);
    ell_A = new double[][] {
      {1, 0, 0},
      {0, 1, 0},
      {0, 0, re_o_rp_2}
    };

  } // EllipsoidModel constructor

  ////////////////////////////////////////////////////////////

  /** Gets the equatorial radius in meters. */
  public double getEquator () { return (rem); }

  ////////////////////////////////////////////////////////////

  /** Gets the polar radius in meters. */
  public double getPolar () { return (rpm); }

  ////////////////////////////////////////////////////////////

  /**
   * Converts from geodetic latitude to geocentric latitude.
   * 
   * @param phiGD the geodetic latitude of the feature in radians.
   * @param height the height above the reference ellipsoid in radians.
   *
   * @return the geocentric latitude in radians.
   */
  public double GDToGCLat (
    double phiGD,
    double height
  ) {

    double cos2_phiGD = Math.pow (Math.cos (phiGD), 2);
    double sin2_phiGD = Math.pow (Math.sin (phiGD), 2);
    double beta =  height*rp*Math.sqrt (re_o_rp_2*cos2_phiGD + sin2_phiGD);
    return (Math.atan (((rp2 + beta)/(re2 + beta)) * Math.tan (phiGD)));

  } // GDToGCLat

  ////////////////////////////////////////////////////////////

  /**
   * Converts from geocentric latitude to geodetic latitude using the
   * height of the feature above the reference ellipsoid.
   * 
   * @param phiGC the geocentric latitude of the feature in radians.
   * @param height the height above the reference ellipsoid in radians.
   *
   * @return the geodetic latitude in radians.
   */
  public double GCToGDLat (
    double phiGC,
    double height
  ) {

    double tan_phiGC = Math.tan (phiGC);
    double phiGD = phiGC;
    double lastPhiGD = 0;
    boolean isStable;
    int count = 0;
    do {

      // Compute new value
      // -----------------
      double cos2_phiGD = Math.pow (Math.cos (phiGD), 2);
      double sin2_phiGD = Math.pow (Math.sin (phiGD), 2);
      double beta = height*rp*Math.sqrt (re_o_rp_2*cos2_phiGD + sin2_phiGD);
      double tan_phiGD = ((re2 + beta)/(rp2 + beta))*tan_phiGC;
      lastPhiGD = phiGD;
      phiGD = Math.atan (tan_phiGD);
      isStable = Math.abs (lastPhiGD - phiGD) < 1e-10;

      // Check convergence limit
      // -----------------------
      count++;
      if (count > MAX_ITERATIONS) {
        throw new RuntimeException ("Latitude did not converge within " +
          "iteration limit");
      } // if
      
    } while (!isStable);

    return (phiGD);

  } // GCToGDLat

  ////////////////////////////////////////////////////////////

  /**
   * Converts from geocentric latitude to geodetic latitude using the
   * distance of the feature from the center of the ellipsoid.
   * 
   * @param phiGC the geocentric latitude of the feature in radians.
   * @param radius the distance from the center of the reference
   * ellipsoid in radians.
   *
   * @return the geodetic latitude in radians.
   */
  public double GCToGDLat2 (
    double phiGC,
    double radius
  ) {

    double phiGD = phiGC;
    double lastPhiGD = 0;
    double height = 0;
    double lastHeight = 0;
    double r2 = radius*radius;
    boolean isStable;
    int count = 0;
    do {

      // Compute new value
      // -----------------
      double cos2_phiGD = Math.pow (Math.cos (phiGD), 2);
      double sin2_phiGD = Math.pow (Math.sin (phiGD), 2);
      double alpha = re_o_rp_2*cos2_phiGD + sin2_phiGD;
      double beta = re_o_rp_4*cos2_phiGD + sin2_phiGD;
      double a = 2*rp*Math.sqrt (alpha);
      double b = rp2*beta/alpha - r2;
      lastHeight = height;
      height = (-a + Math.sqrt (a*a - 4*b))/2;
      lastPhiGD = phiGD;
      phiGD = GCToGDLat (phiGC, height);
      isStable = (
        Math.abs (lastPhiGD - phiGD) < 1e-10 &&
        Math.abs (lastHeight - height) < 1e-10
      );

      // Check convergence limit
      // -----------------------
      count++;
      if (count > MAX_ITERATIONS) {
        throw new RuntimeException ("Latitude did not converge within " +
          "iteration limit");
      } // if
      
    } while (!isStable);

    return (phiGD);

  } // GCToGDLat2

  ////////////////////////////////////////////////////////////

  /**
   * Converts geodetic coordinates to Earth-centered Earth-fixed.
   *
   * @param earthLoc the geodetic Earth location as [lat,lon] in radians.
   * @param height the height above the reference ellipsoid in radians.
   * @param output the output coordinate array to fill.  If null, an
   * output array is created.
   *
   * @return the output coordinate array as [x,y,z] in radians.
   */
  public double[] GDToECEF (
    double[] earthLoc, 
    double height,
    double[] output
  ) {

    if (output == null) output = new double[3];
    double theta = earthLoc[LON];
    double phi = GDToGCLat (earthLoc[LAT], height);
    double[] dirVector = new double[] {
      Math.cos (theta) * Math.cos (phi),
      Math.sin (theta) * Math.cos (phi),
      Math.sin (phi)
    };
    double alpha = VectorOp.dot (dirVector, 
      VectorOp.multiply (ell_A, dirVector, null));
    double t = Math.sqrt (-ELL_C/alpha);
    VectorOp.multiply (dirVector, t + height, output);
    return (output);

  } // GDToECEF

  ////////////////////////////////////////////////////////////

  /**
   * Converts geocentric coordinates to Earth-centered Earth-fixed.
   *
   * @param earthLoc the geocentric Earth location as [lat,lon] in
   * radians.
   * @param radius the location radius in radians.
   * @param output the output coordinate array to fill.  If null, an
   * output array is created.
   *
   * @return the output coordinate array as [x,y,z] in radians.
   */
  public double[] GCToECEF (
    double[] earthLoc, 
    double radius,
    double[] output
  ) {

    if (output == null) output = new double[3];
    double theta = earthLoc[LON];
    double phi = earthLoc[LAT];
    output[X] = radius * Math.cos (theta) * Math.cos (phi);
    output[Y] = radius * Math.sin (theta) * Math.cos (phi);
    output[Z] = radius * Math.sin (phi);
    return (output);

  } // GCToECEF

  ////////////////////////////////////////////////////////////
  
  /**
   * Converts Earth-centered Earth-fixed coordinates to geodetic.
   *
   * @param vector the ECEF coordinate vector as [x,y,z] in radians.
   * @param height the height above the reference ellipsoid in radians.
   * @param output the output Earth location to fill.  If null, an
   * output location is created.
   *
   * @return the output geodetic Earth location as [lat,lon] in radians.
   */
  public double[] ECEFToGD (
    double[] vector,
    double height,
    double[] output
  ) {

    if (output == null) output = new double[2];
    output[LAT] = GCToGDLat (Math.asin (vector[Z] / 
      VectorOp.magnitude (vector)), height);
    output[LON] = Math.atan2 (vector[Y], vector[X]);
    return (output);

  } // ECEFToGD

  ////////////////////////////////////////////////////////////

  /**
   * Computes the normal vector to the ellipsoid surface at the specified
   * surface point.
   *
   * @param posVector the surface position vector as [x,y,z] in radians.
   * @param output the output vector to fill.  If null, an output
   * vector is created.
   *
   * @return the output unit vector as [x,y,z].
   */
  public double[] getNormal (
    double[] posVector,
    double[] output
  ) {

    // Create output array if needed
    // -----------------------------
    if (output == null) output = new double[3];

    // Compute surface normal
    // ----------------------
    output[X] = posVector[X];
    output[Y] = posVector[Y];
    output[Z] = posVector[Z]/rp;
    VectorOp.normalize (output, output);

    return (output);

  } // getNormal

  ////////////////////////////////////////////////////////////

  /**
   * Computes the two solutions to the quadratic equation ax^2 + bx +
   * c = 0.
   *
   * @param a the a coefficient.
   * @param b the b coefficient.
   * @param c the c coefficient.
   * @param output the output array of 2 solutions.
   */
  private static double[] qsolve (
    double a,
    double b,
    double c,
    double[] output
  ) {

    /**
     * Note: This form of the computation helps to reduce
     * roundoff error when sqrt(b^2 - 4ac) is approximately equal
     * to b.  See "Numerical Recipes in C", page 184.
     */
    if (output == null) output = new double[2];
    double q = - (b + (b > 0 ? 1 : -1) * Math.sqrt (b*b - 4*a*c)) / 2;
    output[0] = q/a;
    output[1] = c/q;
    return (output);

  } // qsolve

  ////////////////////////////////////////////////////////////

  /** 
   * Computes the intersection point between a line and the ellipsoid.
   * The line is specified by its starting point and direction.
   *
   * @param posVector the position vector as [x,y,z] in radians.
   * @param dirVector the unit direction vector as [x,y,z].
   * @param output the output intersection array to fill.  If null, an
   * output location is created.
   *
   * @return the output intersection array as [x,y,z] in radians.
   */
  public double[] intersect (
    double[] posVector,
    double[] dirVector,
    double[] output
  ) {

    // Create output array if needed
    // -----------------------------
    if (output == null) output = new double[3];

    // Compute gamma coefficient
    // -------------------------
    if (!Arrays.equals (posVector, lastPosVector)) {
      gamma = ELL_C + VectorOp.dot (posVector, VectorOp.multiply (ell_A, 
        posVector, tempVector));
      for (int i = 0; i < 3; i++) lastPosVector[i] = posVector[i];
    } // if

    // Compute intersection with ellipsoid
    // -----------------------------------
    /**
     * The intersection of the ray and the ellipsoid is computed using
     * a quadratic equation in the parameter t, which is the
     * multiplier of the direction vector: x(t) = e + td.
     */
    double alpha = VectorOp.dot (dirVector, 
      VectorOp.multiply (ell_A, dirVector, tempVector));
    double beta = 2*VectorOp.dot (dirVector, 
      VectorOp.multiply (ell_A, posVector, tempVector));
    qsolve (alpha, beta, gamma, solution);
    double t = Math.min (solution[0], solution[1]);

    // Check for non-intersecting ray
    // ------------------------------
    if (Double.isNaN (t)) {
      for (int i = 0; i < 3; i++) output[i] = Double.NaN;
    } // if

    // Compute final intersection vector
    // ---------------------------------
    else {
      VectorOp.multiply (dirVector, t, output);
      VectorOp.add (posVector, output, output);
    } // else

    return (output);

  } // intersect

  ////////////////////////////////////////////////////////////

} // EllipsoidModel class

////////////////////////////////////////////////////////////////////////
