////////////////////////////////////////////////////////////////////////
/*
     FILE: SensorModel.java
  PURPOSE: Provides sensor model operations.
   AUTHOR: Peter Hollemans
     DATE: 2005/10/14
  CHANGES: n/a

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

// Imports
// -------
import Jama.*;

/**
 * The <code>SensorModel</code> class holds image sensor position and
 * orientation information relative to an {@link EllipsoidModel} and
 * allows for the retrieval of the sensor view location at the surface
 * of the ellipsoid.  The sensor is assumed to be located on a
 * platform that has a known position over the ellipsoid and a nominal
 * pointing direction that is independent of any attitude correction
 * or sensor tilt and rotation.  The platform may have some additional
 * attitude (roll, pitch and yaw) with respect to the nominal axes,
 * and the sensor itself may have some tilt and rotation with respect
 * to the attitude-corrected axes.<p>
 * 
 * The sensor coordinate system has three axes:
 * <ol>
 *
 * <li>The tilt axis about which the sensor can tilt (normally "up" and
 * "down" with respect to the sensor image).</li>
 *
 * <li>The rotation axis about which the sensor can rotate (normally
 * "left" and "right" with respect to the sensor image).</li>
 * 
 * <li>The view axis which defines the direction in which the sensor is
 * pointing.</li>
 *
 * </ol>
 */
public class SensorModel {

  // Constants
  // ---------

  /** The unit vector in the x direction. */
  private static final double[] XHAT = new double[] {1, 0, 0};

  // Variables
  // ---------

  /** The ellipsoid model for the sensor view. */
  private EllipsoidModel viewModel;

  /** 
   * The nominal rotation matrix for platform-fixed to
   * ellipsoid-fixed coordinates.
   */
  private double[][] nominalRotation;

  /** The attitude rotation matrix. */
  private double[][] attitudeRotation;

  /** The platform location vector. */
  private double[] locVector = new double[3];

  /** Various temporary computation arrays. */
  private double[][] ry = new double[3][3];
  private double[][] rz = new double[3][3];
  private double[][] scanRotation = new double[3][3];
  private double[] dirVector = new double[3];
  private double[] dirVectorP = new double[3];

  ////////////////////////////////////////////////////////////

  /** 
   * Creates a new model.
   * 
   * @param model the ellipsoid model to use for the sensor view.
   */
  public SensorModel (
    EllipsoidModel model
  ) {

    viewModel = model;
    locVector = new double[3];

  } // SensorModel constructor

  ////////////////////////////////////////////////////////////

  /**
   * Sets the sensor platform location with respect to the ellipsoid.
   *
   * @param location the location as [x,y,z] in radians.
   */
  public void setLocation (
    double[] location
  ) {

    for (int i = 0; i < 3; i++) locVector[i] = location[i];

  } // setLocation

  ////////////////////////////////////////////////////////////

  /**
   * Sets the nominal sensor platform axes with respect to the
   * ellipsoid.  The axes should be orthogonal.
   *
   * @param viewAxis the direction of the sensor view as an [x,y,z]
   * unit vector.
   * @param tiltAxis the direction of sensor tilt as an [x,y,z] unit
   * vector.
   * @param rotationAxis the direction of sensor rotation as an
   * [x,y,z] unit vector.
   */
  public void setNominalAxes (
    double[] viewAxis,
    double[] tiltAxis,
    double[] rotationAxis
  ) {

    // Create nominal rotation matrix
    // ------------------------------
    double[][] nominalRotationInv = new double[3][];
    nominalRotationInv[VectorOp.X] = viewAxis;
    nominalRotationInv[VectorOp.Y] = tiltAxis;
    nominalRotationInv[VectorOp.Z] = rotationAxis;
    nominalRotation = new Matrix (nominalRotationInv).inverse().getArrayCopy();

  } // setNominalAxes

  ////////////////////////////////////////////////////////////

  /** 
   * Sets the sensor platform attitude angles with respect to the
   * nominal axes.
   *
   * @param roll the roll angle about the sensor rotation axis in
   * radians.
   * @param pitch the pitch angle about the sensor tilt axis in
   * radians.
   * @param yaw the yaw angle about the sensor view axis in radians.
   */
  public void setAttitude (
    double roll,
    double pitch,
    double yaw
  ) {



    // TODO: Construct platform attitude transform.  Look at the
    // KLM user's guide Appendix I for an example.





  } // setAttitude

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the location at which the view axis intersects the ellipsoid
   * surface.
   *
   * @param tilt the sensor tilt angle in radians.
   * @param rotation the sensor rotation angle in radians.
   * @param output the output intersection array to fill.  If null, an
   * output location is created.
   *
   * @return the output intersection array as [x,y,z] in radians.
   */
  public double[] getViewLocation (
    double tilt,
    double rotation,
    double[] output
  ) {

    // Compute scan rotation matrix
    // ----------------------------
    /** 
     * The scan rotation matrix is computed by rotation around the
     * y axis to get the sensor elevation, and then rotation about the
     * z axis for the sensor azimuth.
     */
    VectorOp.rotationMatrix (VectorOp.Y, tilt, ry);
    VectorOp.rotationMatrix (VectorOp.Z, rotation, rz);
    VectorOp.multiply (rz, ry, scanRotation);

    // Compute direction vector
    // ------------------------
    /** 
     * The scan direction is computed by transforming the view
     * direction of the sensor (x axis) using the scan rotation
     * matrix, then transforming to the coordinate system of the
     * ellipsoid.
     */
    VectorOp.multiply (scanRotation, XHAT, dirVectorP);
    VectorOp.multiply (nominalRotation, dirVectorP, dirVector);

    // Get ellipsoid intersection
    // --------------------------
    return (viewModel.intersect (locVector, dirVector, output));

  } // getViewLocation

  ////////////////////////////////////////////////////////////

  /** Tests this class. */
  public static void main (String[] argv) {

    SensorModel model = new SensorModel (EllipsoidModelFactory.getWGS84());
    model.setLocation (new double[] {2, 0, 0});
    model.setNominalAxes (
      new double[] {-1, 0, 0},
      new double[] {0, -1, 0},
      new double[] {0, 0, 1}
    );
    System.out.println ("View location should be [1,0,0]: " + 
      VectorOp.toString (model.getViewLocation (0, 0, null)));
    double cos45 = Math.cos (Math.PI/4);
    double angle = Math.atan (cos45/(2-cos45));
    System.out.println ("Testing with rotation angle = " + 
      Math.toDegrees (angle) + " deg");
    System.out.println ("View location should be [0.707106,-0.707106,0]: " + 
      VectorOp.toString (model.getViewLocation (0, angle, null)));
    System.out.println ("View location should be [0.707106,0.707106,0]: " + 
      VectorOp.toString (model.getViewLocation (0, -angle, null)));

  } // main

  ////////////////////////////////////////////////////////////

} // SensorModel class

////////////////////////////////////////////////////////////////////////



