////////////////////////////////////////////////////////////////////////
/*
     FILE: Spacecraft.java
  PURPOSE: Interface for retrieving spacecraft information.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/20
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

// Imports
// -------
import java.util.*;
import terrenus.instrument.*;

/**
 * The <code>Spacecraft</code> interface delivers information
 * about a spacecraft, its properties, and position/velocity.
 */
public interface Spacecraft {

  ////////////////////////////////////////////////////////////

  /** Gets the spacecraft common name. */
  public String getName();

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the Greenwich hour angle.
   *
   * @param date the date for Greenwich hour angle retrieval.
   *
   * @return the Greenwich hour angle in radians.
   */
  public double getHourAngle (
    Date date 
  );

  ////////////////////////////////////////////////////////////

  /**
   * Gets the position of the spacecraft in earth-centered
   * inertial coordinates.
   * 
   * @param date the date for position retrieval.
   * @param output the output position array to fill.  If null, an
   * output array is created.
   *
   * @return the output position array as [x,y,z] in kilometers.
   */
  public double[] getPosition (
    Date date,
    double[] output
  );

  ////////////////////////////////////////////////////////////

  /**
   * Gets the velocity of the spacecraft in earth-centered
   * inertial coordinates.
   * 
   * @param date the date for velocity retrieval.
   * @param output the output velocity array to fill.  If null, an
   * output array is created.
   *
   * @return the output velocity array as [x,y,z] in
   * kilometers/second.
   */
  public double[] getVelocity (
    Date date,
    double[] output
  );

  ////////////////////////////////////////////////////////////

  /** Gets the spacecraft launch date. */
  public Date getLaunchDate();

  ////////////////////////////////////////////////////////////

  /** Gets the list of known spacecraft instruments. */
  public List<Instrument> getInstruments();

  ////////////////////////////////////////////////////////////

  /**
   * Gets a spacecraft property by name.
   *
   * @param name the name of the property.
   *
   * @return the property value or null if no property with the
   * specified name exists.
   */
  public String getProperty (String name);

  ////////////////////////////////////////////////////////////

} // Spacecraft interface

////////////////////////////////////////////////////////////////////////
