////////////////////////////////////////////////////////////////////////
/*
     FILE: SpacecraftLocator.java
  PURPOSE: Computes spacecraft position and velocity.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/20
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

// Imports
// -------
import java.util.*;
import sgp.*;

/**
 * The <code>SpacecraftLocator</code> class models a spacecraft
 * position and velocity relative to a WGS 72 earth model.
 */
public class SpacecraftLocator {

  // Constants
  // ---------

  /** The UTC timezone. */
  private static final TimeZone UTC = TimeZone.getTimeZone ("GMT+0");

  /** Orbital element time tolerance in days. */
  private static final int ELEMENT_TOLERANCE = 3;

  /** The number of milliseconds per day. */
  private static final int MSEC_PER_DAY = 86400000;

  // Variables
  // ---------

  /** The underlying satellite object from the SGP package. */
  private Satelite sat;

  /** The Greenwich hour angle in radians. */
  private double hourAngle;

  /** The epoch of the orbital elements. */
  private Date epoch;

  ////////////////////////////////////////////////////////////

  /** Gets the orbital element epoch used to initialize this locator. */
  public Date getEpoch () { return (epoch); }

  ////////////////////////////////////////////////////////////

  /** 
   * Determines if a date is within tolerance based on the
   * orbtial element epoch used to create this locator.
   *
   * @param date the date to check.
   *
   * @return true if the date is within tolerance or false if not.
   */
  public boolean isWithinTolerance (
    Date date
  ) {

    double days = (double) (date.getTime() - epoch.getTime()) / MSEC_PER_DAY;
    if (Math.abs (days) < ELEMENT_TOLERANCE) return (true);
    else return (false);

  } // isWithinTolerance

  ////////////////////////////////////////////////////////////

  /** 
   * Initializes this spacecraft using the specified orbital
   * elements. 
   */
  private void initialize (
    String line1,
    String line2
  ) {

    // Create underlying spacecraft object
    // ----------------------------------
    TLE element = new TLE ("", line1, line2);
    sat = new Satelite (element);

    // Save element epoch
    // ------------------
    epoch = AbstractOrbitalElementSource.getEpoch (new String[] {line1, 
      line2});

  } // initialize

  ////////////////////////////////////////////////////////////

  /** 
   * Creates a new spacecraft from a set of two-line orbital
   * elements.
   *
   * @param line1 two-line orbital element line 1.
   * @param line2 two-line orbital element line 2.
   */
  public SpacecraftLocator (
    String line1,
    String line2
  ) {

    initialize (line1, line2);

  } // SpacecraftLocator constructor

  ////////////////////////////////////////////////////////////

  /**
   * Creates a new spacecraft using its NORAD orbital element ID.
   *
   * @param id the NORAD orbital element ID.
   * @param date the date for orbital element retrieval.
   */
  public SpacecraftLocator (
    int id,
    Date date
  ) {

    // Get orbital elements
    // --------------------
    String[] elements = OrbitalElementManager.getInstance().getElements (
      id, date, ELEMENT_TOLERANCE);
    if (elements == null) {
      throw new RuntimeException ("Cannot locate orbital elements within " + 
        ELEMENT_TOLERANCE + " day tolerance");
    } // if

    // Initialize
    // ----------
    initialize (elements[0], elements[1]);

  } // SpacecraftLocator constructor

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the Greenwich hour angle.
   *
   * @return the Greenwich hour angle in radians.
   */
  public double getHourAngle () { return (hourAngle); }

  ////////////////////////////////////////////////////////////

  /** 
   * Sets the date object and performs the spacecraft position and
   * velocity computation.  This method must be called prior to using
   * {@link #getPosition} or {@link #getVelocity}.
   *
   * @param date the date for the spacecraft computations.
   */
  public void setDate (
    Date date
  ) {

    // Compute new position and velocity
    // ---------------------------------
    Calendar targetDate = new GregorianCalendar (UTC);
    targetDate.setTime (date);
    MarcaDeTiempo targetTime = new MarcaDeTiempo();
    targetTime.setValues (
      targetDate.get (Calendar.YEAR),
      targetDate.get (Calendar.MONTH),
      targetDate.get (Calendar.DATE),
      targetDate.get (Calendar.HOUR_OF_DAY),
      targetDate.get (Calendar.MINUTE),
      targetDate.get (Calendar.SECOND),
      targetDate.get (Calendar.MILLISECOND)
    );
    sat.calcularVariables (targetTime);
    
    // Compute Greenwich hour angle
    // ----------------------------
    hourAngle = Tiempo.thetaG_JD (sat.tiempoJuliano);

  } // setDate

  ////////////////////////////////////////////////////////////

  /**
   * Gets the position of the spacecraft in earth-centered
   * inertial coordinates.
   * 
   * @param output the output position array to fill.  If null, an
   * output array is created.
   *
   * @return the output position array as [x,y,z] in kilometers.
   */
  public double[] getPosition (
    double[] output
  ) {
  
    if (output == null) output = new double[3];
    for (int i = 0; i < 3; i++) output[i] = sat.pos[i];
    return (output);

  } // getPosition

  ////////////////////////////////////////////////////////////

  /**
   * Gets the velocity of the spacecraft in earth-centered
   * inertial coordinates.
   * 
   * @param output the output velocity array to fill.  If null, an
   * output array is created.
   *
   * @return the output velocity array as [x,y,z] in
   * kilometers/second.
   */
  public double[] getVelocity (
    double[] output
  ) {
  
    if (output == null) output = new double[3];
    for (int i = 0; i < 3; i++) output[i] = sat.vel[i];
    return (output);

  } // getVelocity

  ////////////////////////////////////////////////////////////

} // SpacecraftLocator class

////////////////////////////////////////////////////////////////////////
