////////////////////////////////////////////////////////////////////////
/*
     FILE: AbstractTelemetryStream.java
  PURPOSE: Implements some parts of telemetry streams.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/11
  CHANGES: 2010/12/18, PFH, removed unneeded cast to int from Math.min
           2010/12/20, PFH, added default day count implementations

  Copyright (C) 2007-2010, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.nio.*;
import java.nio.channels.*;

/**
 * The <code>AbstractTelemetryStream</code> class is a stream that
 * obtains data from a buffer.  The buffer may have extra data at the
 * start and before each frame, and may be in little or big endian
 * byte order for multibyte data streams.
 */
public abstract class AbstractTelemetryStream
  implements TelemetryStream {

  // Variables
  // ---------

  /** The buffer used for data input. */
  protected ByteBuffer input;

  /** The input header length. */
  private int headerLength;

  /** The input frame length. */
  private int frameLength;

  /** The input frame header length. */
  private int frameHeaderLength;

  /** The data start year. */
  private int year = -1;

  /** The data start day. */
  private int dayCount = -1;

  ////////////////////////////////////////////////////////////

  /** 
   * Gets a mapped byte input buffer for the specified data file.
   *
   * @param fileName the name of the file.
   *
   * @return a buffer for accessing the data.
   *
   * @throws IOException if an error occurred accessing the file.
   */
  public static ByteBuffer getInputBuffer (
    String fileName
  ) throws IOException {

    File file = new File (fileName);
    long fileLength = file.length();
    RandomAccessFile raFile = new RandomAccessFile (file, "r");
    FileChannel channel = raFile.getChannel();
    ByteBuffer buffer = channel.map (FileChannel.MapMode.READ_ONLY,
      0, fileLength);

    return (buffer);

  } // getInputBuffer

  ////////////////////////////////////////////////////////////

  /**
   * Initializes a new stream using the specified buffer as input.
   *
   * @param input the input buffer to read.
   * @param headerLength the input header length in bytes.
   * @param frameLength the frame length in bytes (including header).
   * @param frameHeaderLength the frame header length.
   * @param isLittleEndian the little endian flag, true for LSB first or
   * false for MSB first.
   */
  protected void initStream (
    ByteBuffer input,
    int headerLength,
    int frameLength,
    int frameHeaderLength,
    boolean isLittleEndian
  ) {

    // Initialize
    // ----------
    this.input = input;
    this.headerLength = headerLength;
    this.frameLength = frameLength;
    this.frameHeaderLength = frameHeaderLength;
    if (isLittleEndian) input.order (ByteOrder.LITTLE_ENDIAN);

    // Check stream
    // ------------
    if (!checkStream())
      throw new RuntimeException ("Cannot find valid frame data");

  } // initStream

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the position of a frame within the input.
   * 
   * @param frame the frame in question.
   *
   * @return the position in bytes, relative to the end of any input
   * header data.
   */
  protected int getFramePosition (
    int frame
  ) {

    return (frame*frameLength);

  } // getFramePosition

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the position of a block within a frame.
   * 
   * @param block the block in question.
   *
   * @return the position in bytes, relative to the end of any frame
   * header data.
   */
  protected abstract int getBlockPosition (
    int block
  );

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the length of a block.
   *
   * @param block the block in question.
   *
   * @return the length of the block in bytes.
   */
  protected abstract int getBlockLength (
    int block
  );

  ////////////////////////////////////////////////////////////

  public ByteBuffer getBlock (
    int frame,
    int block
  ) {

    int offset = headerLength + getFramePosition (frame) + 
      frameHeaderLength + getBlockPosition (block);
    input.clear().position (offset);
    input.limit (input.position() + getBlockLength (block));
    return (input.slice().order (input.order()));

  } // getBlock

  ////////////////////////////////////////////////////////////

  public int getYear () throws IOException { return (year); }

  ////////////////////////////////////////////////////////////

  public void setYear (int year) { this.year = year; }

  ////////////////////////////////////////////////////////////

  public int getDayCount () throws IOException { return (dayCount); }

  ////////////////////////////////////////////////////////////

  public void setDayCount (int dayCount) { this.dayCount = dayCount; }

  ////////////////////////////////////////////////////////////

  /**
   * Checks a frame for errors.  The check procedure may involve
   * looking for expected patterns such as synchronization bytes.
   *
   * @param frame the frame number to check.
   *
   * @return true if the frame check was successful, or false if
   * it failed.
   *
   * @see #checkStream
   */
  protected abstract boolean checkFrame (
    int frame
  );

  ////////////////////////////////////////////////////////////

  /**
   * Checks the telemetry stream for errors.  This method calls
   * {@link #checkFrame} for a contiguous sequence of frames at
   * the start of the stream to determine if the stream has valid
   * frame data.  Child classes may want to override this method
   * to perform more elaborate stream checking.
   *
   * @return true if the frame check was successful, or false if
   * it failed.
   */
  protected boolean checkStream () {

    // Set frame check parameters
    // --------------------------
    int maxFrames = Math.min (getFrameCount(), 500);
    int minValid = maxFrames/10;
    int validCount = 0;

    // Check starting frames
    // ---------------------
    for (int frame = 0; frame < maxFrames && validCount < minValid; frame++) {
      if (!checkFrame (frame)) validCount--;
      else validCount++;
    } // for

    // Check valid count
    // -----------------
    if (validCount < minValid) return (false);
    else return (true);

  } // checkStream

  ////////////////////////////////////////////////////////////

  public int getFrameCount () { 

    return ((input.capacity() - headerLength)/frameLength);

  } // getFrameCount

  ////////////////////////////////////////////////////////////

  /** Tests this class. */
  public static void main (String[] argv) throws Exception {

    int i = 0;
    String name = argv[i++];
    TelemetryStream stream = TelemetryStreamFactory.create (name);

    int startFrame = Integer.parseInt (argv[i++]);
    int endFrame = Integer.parseInt (argv[i++]);
    int blocks = Integer.parseInt (argv[i++]);
    for (int frame = startFrame; frame <= endFrame; frame++) {
      System.out.print (frame + " ");
      for (int block = 0; block < blocks; block++) {
        ShortBuffer buf = stream.getBlock (frame, block).asShortBuffer();
        while (buf.hasRemaining()) System.out.print (buf.get() + " ");
      } // for
      System.out.println();
    } // for

  } // main

  ////////////////////////////////////////////////////////////

} // AbstractTelemetryStream class

////////////////////////////////////////////////////////////////////////
