////////////////////////////////////////////////////////////////////////
/*
     FILE: CHRPTDecoder.java
  PURPOSE: Decodes CHRPT data.
   AUTHOR: Peter Hollemans
           Mohammed Kamal
     DATE: 2007/02/08
  CHANGES: 2007/03/16, PFH, modified to implement TelemetryDecoder
           2010/12/18, PFH, removed unneeded cast to int from millisecond |= ...

  Copyright (C) 2007-2010, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.util.*;
import java.nio.*;
import java.net.*;
import terrenus.instrument.*;
import terrenus.instrument.RadiometerCalibrator.Record;
import terrenus.orbit.*;
import terrenus.util.*;

/**
 * The <code>CHRPTDecoder</code> class uses a {@link CHRPTStream} to
 * decode CMA CHRPT data.
 */
public class CHRPTDecoder
  implements TelemetryDecoder {

  // Constants
  // ---------

  /** The mask for bit 1 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT1 = (short) 0x0200;

  /** The mask for bit 2 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT2 = (short) 0x0100;

  /** The mask for bit 3 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT3 = (short) 0x0080;

  /** The mask for bit 4 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT4 = (short) 0x0040;

  /** The mask for bit 5 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT5 = (short) 0x0020;

  /** The mask for bit 6 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT6 = (short) 0x0010;

  /** The mask for bit 7 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT7 = (short) 0x0008;

  /** The mask for bit 8 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT8 = (short) 0x0004;

  /** The mask for bit 9 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT9 = (short) 0x0002;

  /** The mask for bit 10 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT10 = (short) 0x0001;

  /** The spacecraft address for FY-1D. */
  private static final int FY1D_ADDRESS = 13;

  /** The number of milliseconds between adjacent frames. */
  private static final double MSEC_PER_FRAME = 1000.0/6;

  /** The number of milliseconds in one hour. */
  private static final int MSEC_PER_HOUR = 3600000;

  /** The number of milliseconds in one minute. */
  private static final int MSEC_PER_MINUTE = 60000;

  /** The number of milliseconds in one second. */
  private static final int MSEC_PER_SECOND = 1000;

  // Variables
  // ---------

  /** The stream for obtaining data. */
  private CHRPTStream stream;

  /** The actual spacecraft address. */
  private int spacecraftId;

  /** The spacecraft corresponding to the spacecraft address. */
  private Spacecraft spacecraft;

  /** The actual data start time in milliseconds since the epoch. */
  private long startTime;

  /** The array of frame sync errors. */
  private int[] frameSyncErrors;

  /** The array of auxiliary sync errors. */
  private int[] auxSyncErrors;

  /** The calibrator for MVISR calibration. */
  private MVISRCalibrator mvisrCalibrator;

  /** The locator for MVISR location. */
  private MVISRLocator mvisrLocator;

  ////////////////////////////////////////////////////////////

  static {

    /**
     * To aid in using CMA CHRPT data, we add various sources of
     * orbital elements here.  That way, other classes can make use of
     * them through the {@link OrbitalElementManager} without having
     * to know where they came from.
     */
    try {

      // Add archive sources
      // -------------------
      OrbitalElementManager manager = OrbitalElementManager.getInstance();
      manager.registerSource (new ArchiveElementSource (
        OrbitalElementManager.class.getResource ("fy1d.txt")));

      // Add current sources
      // -------------------
      manager.registerSource (new CurrentElementSource (
        new URL ("http://www.celestrak.net/NORAD/elements/weather.txt")));

    } // try
    catch (IOException e) {
      throw new RuntimeException ("Error loading default orbital elements");
    } // catch

  } // static

  ////////////////////////////////////////////////////////////

  public boolean hasErrors (int frame) throws IOException {

    boolean errors = (
      frameSyncErrors[frame] != 0 ||
      auxSyncErrors[frame] != 0
    );
    return (errors);

  } // hasErrors

  ////////////////////////////////////////////////////////////

  /** Create a new decoder using the specified stream. */
  public CHRPTDecoder (CHRPTStream stream) throws IOException { 

    this.stream = stream; 

    // Create frame error arrays
    // -------------------------
    int frameCount = stream.getFrameCount();
    frameSyncErrors = new int[frameCount];
    auxSyncErrors = new int[frameCount];

    // Initialize statistical tests
    // ----------------------------
    FrequencyMap<Integer> spacecraftMap = new FrequencyMap<Integer>();
    ArrayList<Long> startTimeList = new ArrayList<Long>();
    Calendar cal = new GregorianCalendar (TimeZone.getTimeZone ("GMT+0"));
    int year = stream.getYear();
    if (year == -1) 
      throw new RuntimeException ("Cannot determine year from CHRPT data");

    // Loop over each frame
    // --------------------
    for (int i = 0; i < frameCount; i++) {

      // Compute bit errors
      // ------------------
      frameSyncErrors[i] = computeFrameSyncErrors (i);
      auxSyncErrors[i] = computeAuxSyncErrors (i);

      // Check if data is usable
      // -----------------------
      boolean isUsable = !hasErrors (i);
      if (!isUsable) continue;

      // Add spacecraft to frequency map
      // -------------------------------
      spacecraftMap.put (getSpacecraftId (i));

      // Add start time to list
      // ----------------------
      cal.set (Calendar.YEAR, year);
      cal.set (Calendar.DAY_OF_YEAR, getDayCount (i));
      int millisecond = getMillisecond (i);
      int hour = millisecond/MSEC_PER_HOUR;
      cal.set (Calendar.HOUR_OF_DAY, hour);
      int minute = (millisecond - hour*MSEC_PER_HOUR)/MSEC_PER_MINUTE;
      cal.set (Calendar.MINUTE, minute);
      int second = (millisecond - hour*MSEC_PER_HOUR - 
        minute*MSEC_PER_MINUTE)/MSEC_PER_SECOND;
      cal.set (Calendar.SECOND, second);
      int newMilli = millisecond - hour*MSEC_PER_HOUR - minute*MSEC_PER_MINUTE 
        - second*MSEC_PER_SECOND;
      cal.set (Calendar.MILLISECOND, newMilli);
      long thisStartTime = (long) (cal.getTimeInMillis() - i*MSEC_PER_FRAME);
      startTimeList.add (thisStartTime);

    } // for

    // Set spacecraft ID
    // -----------------
    Integer freqValue = spacecraftMap.getFreqValue();
    if (freqValue == null)
      throw new IOException ("Cannot locate valid CHRPT data");
    spacecraftId = freqValue;

    // Set start time
    // --------------
    Collections.sort (startTimeList);
    int median = startTimeList.size()/2;
    startTime = startTimeList.get (median);

  } // CHRPTDecoder

  ////////////////////////////////////////////////////////////

  public CHRPTStream getStream () { return (stream); }

  ////////////////////////////////////////////////////////////

  /** Computes the number of errors in the frame sync bits. */
  private int computeFrameSyncErrors (int frame) throws IOException {

    ByteBuffer buf = stream.getBlock (frame, CHRPTStream.FRAME_SYNC_BLOCK);
    return (HRPTFrameSyncChecker.getInstance().getDifferenceCount (buf));
    
  } // computeFrameSyncErrors

  ////////////////////////////////////////////////////////////

  /**
   * Gets the number of errors in the frame sync bits, or 0 if no
   * errors are present.  The number of errors is the total number of
   * bits which differ between the expected frame sync bits and the
   * actual frame sync bits.
   */
  public int getFrameSyncErrors (int frame) { 

    return (frameSyncErrors[frame]); 

  } // getFrameSyncErrors

  ////////////////////////////////////////////////////////////

  /**
   * Gets the start time of frame capture in milliseconds since the
   * epoch.  This time may differ from the values returned by {@link
   * #getDayCount} and {@link #getMillisecond} since some frames may
   * have bit errors.  The actual frame time is computed by examining
   * a number of frame times, inferring the correct data start time
   * from those times, and then computing the frame time using the
   * frame index and the known frame step interval.  This presents the
   * CHRPT data as a continuous, uniformly stepping set of frames and
   * is very useful for computing correct Earth locations for every
   * sensor scan even when the frame itself contains unusable sensor
   * data.
   *
   * @param frame the frame to return the time for, starting at 0.
   *
   * @return the frame start time in milliseconds since the epoch.
   * The epoch in this case is January 1, 1970 00:00:00 UTC.
   */
  public long getTime (int frame) {

    return ((long) (startTime + frame*MSEC_PER_FRAME));

  } // getTime

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the spacecraft address identified for this CHRPT data.  This
   * may differ from the spacecraft address returned by {@link
   * #getSpacecraftId(int)} since some frames may have bit errors.
   * The actual spacecraft address is detected by examining a number
   * of CHRPT minor frames.
   *
   * @return the actual spacecraft address for this data.
   */
  public int getSpacecraftId () { return (spacecraftId); }

  ////////////////////////////////////////////////////////////

  /** Gets the spacecraft address for the frame. */
  public int getSpacecraftId (int frame) throws IOException {

    ShortBuffer buf = 
      stream.getBlock (frame, CHRPTStream.ID_BLOCK).asShortBuffer();
    return ((buf.get(0) & (BIT1|BIT2|BIT3|BIT4)) >>> 6);

  } // getSpacecraftId

  ////////////////////////////////////////////////////////////

  /** Gets the frame day count. */
  public int getDayCount (int frame) throws IOException {

    /**
     * The day count would normally be stored in the time code
     * words and be simple to extract, but CHRPT data has zeros
     * in the time words where the day count is supposed to be.
     * So we have to rely on getting the day count from some other
     * source that the stream knows about.  We just pass on the
     * information from the stream here.
     */
   return (stream.getDayCount (frame)); 

  } // getDayCount

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the millisecond based on CHRPT time code data.
   *
   * @param timeCode the time code data as an array of four short
   * values.
   *
   * @return the time in milliseconds from the time code data.
   */
  public static int getMillisecond (
    short[] timeCode
  ) {

    int millisecond = 0;
    millisecond |= (timeCode[1] & 0x007f) << 20;
    millisecond |= ((int) timeCode[2]) << 10;
    millisecond |= timeCode[3];
    return (millisecond);

  } // getMillisecond  

  ////////////////////////////////////////////////////////////

  /** Gets the frame millisecond. */
  public int getMillisecond (int frame) throws IOException {

    ShortBuffer buf = 
      stream.getBlock (frame, CHRPTStream.TIME_CODE_BLOCK).asShortBuffer();
    return (getMillisecond (new short[] {0, buf.get(1), buf.get(2), 
      buf.get(3)}));

  } // getMillisecond

  ////////////////////////////////////////////////////////////

  /** Computes the number of errors in the frame auxiliary sync bits. */
  private int computeAuxSyncErrors (int frame) throws IOException {

    ByteBuffer buf = stream.getBlock (frame, CHRPTStream.AUX_SYNC_BLOCK);
    return (HRPTAuxSyncChecker.getInstance().getDifferenceCount (buf));

  } // computeAuxSyncErrors

  ////////////////////////////////////////////////////////////

  /**
   * Gets the number of errors in the frame auxiliary sync bits, or
   * 0 if no errors are present.  The number of errors is the total
   * number of bits which differ between the expected auxiliary sync
   * bits and the actual auxiliary sync bits.
   */
  public int getAuxSyncErrors (int frame) {

    return (auxSyncErrors[frame]);

  } // getAuxSyncErrors

  ////////////////////////////////////////////////////////////

  public InstrumentData getData (
    int frame,
    Instrument inst
  ) throws IOException {

    // Check instrument
    // ----------------
    if (!(inst instanceof MVISR)) 
      throw new IllegalArgumentException ("Illegal instrument: " + 
        inst.getClass().getName());

    // Create MVISR objects
    // --------------------
    if (mvisrCalibrator == null) mvisrCalibrator = new MVISRCalibrator (this);
    if (mvisrLocator == null) mvisrLocator = new MVISRLocator (this);

    // Create MVISR data for frame
    // ---------------------------
    Record record = mvisrCalibrator.getCalibration (frame);
    MVISRData data = new MVISRData (frame, this, mvisrCalibrator, 
      record, mvisrLocator);

    return (data);

  } // getData

  ////////////////////////////////////////////////////////////

  public Spacecraft getSpacecraft () { 

    // Create new spacecraft object
    // ----------------------------
    if (spacecraft == null) {
      int noradId;
      switch (spacecraftId) {
      case FY1D_ADDRESS: noradId = 27431; break;
      default: throw new RuntimeException ("Unknown spacecraft address");
      } // switch
      spacecraft = new FY1Spacecraft (noradId);
    } // if

    return (spacecraft);

  } // getSpacecraft

  ////////////////////////////////////////////////////////////

} // CHRPTDecoder class

////////////////////////////////////////////////////////////////////////
