////////////////////////////////////////////////////////////////////////
/*
     FILE: QuorumCHRPTStream.java
  PURPOSE: Reads Quorum CHRPT data.
   AUTHOR: Peter Hollemans
           Mohammed Kamal
     DATE: 2007/02/08
  CHANGES: 2007/03/16, PFH, modified to extend CHRPTStream

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.util.*;
import terrenus.orbit.*;

/**
 * The <code>QuorumCHRPTStream</code> class reads Quorum QTrack CMA
 * CHRPT storage format.  The file structure is as follows:
 * <ul>
 *   <li>Possible header of unknown size</li>
 *   <li>CHRPT pseudo frames signalled by a 0x8000 start word</li>
 *   <li>44360 bytes per CHRPT frame</li>
 *   <li>10-bit words are stored as least significant 10 bits
 *   in each 16-bit byte-swapped word
 * </ul>
 */
public class QuorumCHRPTStream 
  extends CHRPTStream {

  // Constants
  // ---------

  /** The number of words in one frame. */
  private static final int FRAME_WORDS = 22180;

  // Variables
  // ---------

  /** The Quorum file to use for frame positions and info. */
  private QuorumFile qFile;

  /** The starting day count. */
  private int startDay;

  ////////////////////////////////////////////////////////////

  public String getFormat () { return ("Quorum QTrack CHRPT"); }

  ////////////////////////////////////////////////////////////

  /** Creates a new stream using the specified file as input. */
  public QuorumCHRPTStream (
    String fileName
  ) throws IOException {

    // Open Quorum file and initialize stream
    // --------------------------------------
    qFile = new QuorumFile (fileName, FRAME_WORDS);
    initStream (getInputBuffer (fileName), 0, 0, 0, true);

    // Set stream year
    // ---------------
    Calendar cal = new GregorianCalendar(TimeZone.getTimeZone ("GMT+0"));
    cal.setTimeInMillis (qFile.getStartTime());
    setYear (cal.get (Calendar.YEAR));
    
    // Set stream starting day
    // -----------------------
    startDay = cal.get (Calendar.DAY_OF_YEAR);

    // Register emphemeris with manager
    // --------------------------------
    OrbitalElementManager.getInstance().registerSource (qFile);

  } // QuorumCHRPTStream constructor

  ////////////////////////////////////////////////////////////

  protected int getFramePosition (
    int frame
  ) {

    return (qFile.getFramePosition (frame)*2);

  } // getFramePosition

  ////////////////////////////////////////////////////////////

  public int getFrameCount () { return (qFile.getFrameCount()); }

  ////////////////////////////////////////////////////////////

  public int getDayCount (int frame) throws IOException {

    // TODO: We return the starting day count here and hope
    // that the decoder is smart enough to handle a day change
    // in the middle of a pass.  There should be a better algorithm
    // for this.

    return (startDay);

  } // getDayCount

  ////////////////////////////////////////////////////////////

} // QuorumCHRPTStream class

////////////////////////////////////////////////////////////////////////
