////////////////////////////////////////////////////////////////////////
/*
     FILE: QuorumHRPTStream.java
  PURPOSE: Reads Quorum HRPT data.
   AUTHOR: Peter Hollemans
     DATE: 2005/08/01
  CHANGES: 2007/01/27, PFH, moved main functionality to abstract parent
           2007/01/28, PFH, added checkFormat() and getFormat()
           2007/02/29, PFH, modified to extend HRPTStream

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.util.*;
import terrenus.orbit.*;

/**
 * The <code>QuorumHRPTStream</code> class reads Quorum QTrack NOAA
 * HRPT storage format.  The file structure is as follows:
 * <ul>
 *   <li>Possible header of unknown size</li>
 *   <li>HRPT pseudo frames signalled by a 0x8000 start word</li>
 *   <li>22180 bytes per HRPT minor frame</li>
 *   <li>10-bit words are stored as least significant 10 bits
 *   in each 16-bit little endian value
 * </ul>
 */
public class QuorumHRPTStream 
  extends HRPTStream {

  // Constants
  // ---------

  /** The number of words in one frame. */
  private static final int FRAME_WORDS = 11090;

  // Variables
  // ---------

  /** The Quorum file to use for frame positions and info. */
  private QuorumFile qFile;

  ////////////////////////////////////////////////////////////

  public String getFormat () { return ("Quorum QTrack HRPT"); }

  ////////////////////////////////////////////////////////////

  /** Creates a new stream using the specified file as input. */
  public QuorumHRPTStream (
    String fileName
  ) throws IOException {

    // Open Quorum file and initialize stream
    // --------------------------------------
    qFile = new QuorumFile (fileName, FRAME_WORDS);
    initStream (getInputBuffer (fileName), 0, 0, 0, true);

    // Set stream year
    // ---------------
    Calendar cal = new GregorianCalendar(TimeZone.getTimeZone ("GMT+0"));
    cal.setTimeInMillis (qFile.getStartTime());
    setYear (cal.get (Calendar.YEAR));

    // Register emphemeris with manager
    // --------------------------------
    OrbitalElementManager.getInstance().registerSource (qFile);

  } // QuorumHRPTStream constructor

  ////////////////////////////////////////////////////////////

  protected int getFramePosition (
    int frame
  ) {

    return (qFile.getFramePosition (frame)*2);

  } // getFramePosition

  ////////////////////////////////////////////////////////////

  public int getFrameCount () { return (qFile.getFrameCount()); }

  ////////////////////////////////////////////////////////////

} // QuorumHRPTStream class

////////////////////////////////////////////////////////////////////////
