////////////////////////////////////////////////////////////////////////
/*
     FILE: TelemetryDecoder.java
  PURPOSE: Interface for decoding generic telemetry streams.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/19
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import terrenus.instrument.*;
import terrenus.orbit.*;

/**
 * The <code>TelemetryDecoder</code> class uses a {@link
 * TelemetryStream} to decode telemetry data from a spacecraft
 * telemetry data stream.  The decoder presents a high-level interface
 * for telemetry data, allowing you to make queries on various useful
 * telemetry frame information without the having to resort to
 * manipulating bit fields.  The methods of the enclosed {@link
 * TelemetryStream} may still be used to access the raw data.
 */
public interface TelemetryDecoder {

  ////////////////////////////////////////////////////////////

  /** Gets the telemetry stream used for data. */
  public TelemetryStream getStream();

  ////////////////////////////////////////////////////////////

  /**
   * Gets the start time of frame capture.
   *
   * @param frame the frame to return the time for in the range
   * [0..FRAMES-1].
   *
   * @return the frame start time in milliseconds since the epoch.
   * The epoch in this case is January 1, 1970 00:00:00 UTC.
   */
  public long getTime (int frame);

  ////////////////////////////////////////////////////////////
  
  /**
   * Gets the frame error flag.
   *
   * @param frame the frame to return the flag for in the range
   * [0..FRAMES-1].
   *
   * @return the frame error flag, true if errors were detected in the
   * frame data or false if not.  Error detection usually consists of
   * checking frame synchronization data and bit parity.
   *
   * @throws IOException if an error occurred reading data from the
   * stream.
   */
  public boolean hasErrors (int frame) throws IOException;

  ////////////////////////////////////////////////////////////

  /** Gets the spacecraft detected by this decoder. */
  public Spacecraft getSpacecraft();

  ////////////////////////////////////////////////////////////

  /**
   * Gets the frame data for a specific instrument.
   *
   * @param frame the frame to return data for in the range
   * [0..FRAMES-1].
   * @param inst the instrument to return data for.
   *
   * @return the frame data for the specified instrument.
   *
   * @throws IOException if an error occurred reading the instrument
   * data from the stream.
   */
  public InstrumentData getData (
    int frame,
    Instrument inst
  ) throws IOException;

  ////////////////////////////////////////////////////////////

} // TelemetryDecoder interface

////////////////////////////////////////////////////////////////////////
