////////////////////////////////////////////////////////////////////////
/*
     FILE: TelemetryStream.java
  PURPOSE: Interface for reading generic telemetry streams.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/10
  CHANGES: 2010/12/20, PFH, added day count methods

  Copyright (C) 2007-2010, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.nio.*;

/**
 * A <code>TelemetryStream</code> is a sequence of data values
 * broadcast by a spacecraft.  Telemetry data consists of a series of
 * data frames of some fixed length.  Each frame contains a number of
 * blocks of data.  A stream object provides a decoder that allows for
 * useful operations on the data.
 */
public interface TelemetryStream {

  ////////////////////////////////////////////////////////////

  /** Gets the total number of telemetry frames. */
  public int getFrameCount ();

  ////////////////////////////////////////////////////////////

  /** Gets the starting year, or -1 if not available. */
  public int getYear () throws IOException;

  ////////////////////////////////////////////////////////////

  /** 
   * Sets the starting year.  This may be needed if the year
   * cannot be detected from the stream.
   *
   * @param year the starting year.  The year is returned by the next
   * call to {@link #getYear}.
   */
  public void setYear (int year);

  ////////////////////////////////////////////////////////////

  /** Gets the starting day count, or -1 if not available. */
  public int getDayCount () throws IOException;

  ////////////////////////////////////////////////////////////

  /** 
   * Sets the starting day count.  This may be needed if the day count
   * cannot be detected from the stream.
   *
   * @param dayCount the starting day count.  The day count is returned by the 
   * next call to {@link #getDayCount}.
   */
  public void setDayCount (int dayCount);

  ////////////////////////////////////////////////////////////

  /** Gets a description of the underlying data format. */
  public String getFormat();

  ////////////////////////////////////////////////////////////

  /** 
   * Gets a decoder for the stream data.
   *
   * @throws IOException if an error occurred decoding the stream
   * data.
   */
  public TelemetryDecoder getDecoder() throws IOException;

  ////////////////////////////////////////////////////////////

  /** Gets the number of blocks per frame. */
  public abstract int getBlockCount();

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the data for the specified block.
   *
   * @param frame the frame to get data.
   * @param block the block number to get data.
   *
   * @return the block data buffer.
   */
  public ByteBuffer getBlock (
    int frame,
    int block
  );

  ////////////////////////////////////////////////////////////

} // TelemetryStream interface

////////////////////////////////////////////////////////////////////////
