////////////////////////////////////////////////////////////////////////
/*
     File: HDFServices.java
  PURPOSE: Aids in writing HDF data.
   AUTHOR: Peter Hollemans
     DATE: 2007/03/13
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.util;

// Imports
// -------
import java.io.*;
import java.lang.reflect.*;
import ncsa.hdf.hdflib.*;

/**
 * The <code>HDFServices</code> class contains a number of static
 * methods to help write Hierarchical Data Format (HDF) version 4
 * data files.
 */
public class HDFServices {

  ////////////////////////////////////////////////////////////

  /** 
   * Writes an HDF attribute.
   * 
   * @param sdid the scientific dataset ID for either a file or variable.
   * @param name the name of the attribute.
   * @param value the value for the attribute.  Primitive Java
   * types must be passed as an array.
   *
   * @throws IOException if an HDF or other error occurred
   * writing the attribute.
   */
  public static void writeAtt (
    int sdid,
    String name,
    Object value
  ) throws IOException {

    // Get HDF type
    // ------------
    int hdfType;
    if (value instanceof String) {
      value = ((String)value).getBytes();
      hdfType = HDFConstants.DFNT_CHAR8;
    } // if
    else if (value instanceof byte[]) hdfType = HDFConstants.DFNT_INT8;
    else if (value instanceof short[]) hdfType = HDFConstants.DFNT_INT16;
    else if (value instanceof int[]) hdfType = HDFConstants.DFNT_INT32;
    else if (value instanceof long[]) hdfType = HDFConstants.DFNT_INT64;
    else if (value instanceof float[]) hdfType = HDFConstants.DFNT_FLOAT32;
    else if (value instanceof double[]) hdfType = HDFConstants.DFNT_FLOAT64;
    else throw new IOException ("Cannot find valid HDF type");

    // Write value
    // -----------
    int len = Array.getLength (value);
    try { 
      if (!HDFLibrary.SDsetattr (sdid, name, hdfType, len, value))
        throw new HDFException();
    } // try
    catch (HDFException e) {
      throw new IOException ("Error writing attribute value");
    } // catch

  } // writeAtt

  ////////////////////////////////////////////////////////////

  /** 
   * Creates a new HDF variable with standard attributes.
   *
   * @param sdid the scientific dataset ID for the HDF file.
   * @param name the new variable name to create.
   * @param hdfType the HDF type for the variable.  Currently
   * only DFNT_INT16 (scale_factor = 0.01) and DFNT_FLOAT32 (no
   * scale_factor) are supported.
   * @param units the data units string.
   * @param dims the variable dimensions.
   * @param keepOpen the keep open flag, true to keep the
   * variable open after creating it or false to end access to
   * the variable.
   *
   * @return the scientific dataset ID for the new variable if
   * keepOpen is true, or the SD index if keepOpen is false.
   *
   * @throws IOException if an HDF or other error occurred
   * creating the variable.
   */
  public static int createVariable (
    int sdid,
    String name,
    int hdfType,
    String units,
    int[] dims,
    boolean keepOpen
  ) throws IOException {

    // Create variable
    // ---------------
    int sdsid;
    try {
      sdsid = HDFLibrary.SDcreate (sdid, name, hdfType, dims.length, dims);
      if (sdsid < 0) throw new HDFException();
    } // try
    catch (HDFException e) {
      throw new IOException ("Error creating variable");
    } // catch

    // Set attributes
    // --------------
    try {
      if (!HDFLibrary.SDsetdatastrs (sdsid, name, units, "", ""))
        throw new HDFException();
    } // try
    catch (HDFException e) {
      throw new IOException ("Error writing data strings");
    } // catch
    Object missing;
    double[] scaling;
    if (hdfType == HDFConstants.DFNT_INT16) {
      missing = new short[] {Short.MIN_VALUE};
      scaling = new double[] {0.01, 0};
      writeAtt (sdsid, "fraction_digits", new int[] {2});
    } // if
    else if (hdfType == HDFConstants.DFNT_FLOAT32) {
      missing = new float[] {Float.NaN};
      scaling = null;
      writeAtt (sdsid, "fraction_digits", new int[] {6});
    } // else if
    else 
      throw new IOException ("Invalid HDF variable type");
    try {
      if (!HDFLibrary.SDsetfillvalue (sdsid, missing))
        throw new HDFException();
    } // try
    catch (HDFException e) {
      throw new IOException ("Error setting fill value");
    } // catch
    writeAtt (sdsid, "missing_value", missing);
    if (scaling != null) {
      try {
        if (!HDFLibrary.SDsetcal (sdsid, scaling[0], 0, scaling[1], 0, 0))
          throw new HDFException();
      } // try
      catch (HDFException e) {
        throw new IOException ("Error setting calibration");
      } // catch
    } // if
    
    // End variable access
    // -------------------
    if (!keepOpen) {
      try { 
        HDFLibrary.SDendaccess (sdsid); 
        int index  = HDFLibrary.SDnametoindex (sdid, name);
        return (index);
      } // try
      catch (HDFException e) { 
        throw new IOException ("Error ending variable access");
      } // catch
    } // if

    return (sdsid);

  } // createVariable

  ////////////////////////////////////////////////////////////

  /**
   * Writes data to an HDF variable.
   *
   * @param sdid the scientific dataset ID for the file.
   * @param varRef the variable reference, either the scientific
   * dataset ID for open variables, or the scientific dataset
   * index for closed variables.  See the {@link #createVariable}
   * method for an explanation of open versus closed.
   * @param isOpen the open flag, true if the variable is open
   * already or false if not.  If the variable is not open, it is
   * opened, written to, and then closed.  If the variable is
   * already open, it is left open.
   * @param start the starting data coordinates for writing data.
   * @param stride the stride for for writing data.
   * @param count the count of data values to write in each data
   * dimension.
   * @param data the data value array to write.
   *
   * @throws IOException if an HDF error occurred writing the
   * variable data.
   */
  public static void writeVariable (
    int sdid,
    int varRef,
    boolean isOpen,
    int[] start,
    int[] stride,
    int[] count,
    Object data
  ) throws IOException {

    try {

      // Get SDS ID
      // ----------
      int sdsid;
      if (!isOpen) sdsid = HDFLibrary.SDselect (sdid, varRef);
      else sdsid = varRef;

      // Write data
      // ----------
      if (!HDFLibrary.SDwritedata (sdsid, start, stride, count, data))
        throw new HDFException();

      // End data access
      // ---------------
      if (!isOpen) HDFLibrary.SDendaccess (sdsid);          

    } // try
    catch (HDFException e) {
      throw new IOException ("Error writing variable data");
    } // catch

  } // writeVariable

  ////////////////////////////////////////////////////////////

} // HDFServices class

////////////////////////////////////////////////////////////////////////
